# =============== Modbus Server - Simulation Server ===============
#
# pymodbus TCP server 封裝 + tick loop

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

from csp_lib.core import AsyncLifecycleMixin, get_logger
from csp_lib.modbus._pymodbus import (
    ensure_pymodbus_server,
    get_BaseModbusDataBlock,
    get_ModbusDeviceContext,
    get_ModbusServerContext,
    get_ModbusTcpServer,
)

from .config import ServerConfig

if TYPE_CHECKING:
    from .microgrid import MicrogridSimulator
    from .simulator.base import BaseDeviceSimulator

logger = get_logger(__name__)


def _ensure_pymodbus_imported() -> None:
    """確保 pymodbus server 相關模組已載入 (backward-compat wrapper)."""
    ensure_pymodbus_server()


class SimulatorDataBlock:
    """
    自訂 Modbus DataBlock

    橋接 pymodbus datastore 與 RegisterBlock。
    寫入時偵測 writable points 並呼叫 simulator.on_write()。

    pymodbus 3.12 的 BaseModbusDataBlock 要求 values, address, default_value 屬性。
    透過 _create_datablock() 工廠函式建立繼承版本。
    """

    def __init__(self, simulator: BaseDeviceSimulator) -> None:
        self._simulator = simulator
        self._block = simulator.register_block
        # pymodbus 需要這些屬性
        self.address = 0
        self.default_value = 0
        self.values = self._block.registers

    def reset(self) -> None:
        pass

    def getValues(self, address: int, count: int = 1) -> list[int]:
        return self._block.get_raw(address, count)

    def setValues(self, address: int, values: list[int]) -> None:
        # 先記錄受影響 points 的舊值
        affected = self._block.find_affected_points(address, len(values))
        old_values = {}
        for point in affected:
            old_values[point.name] = self._block.get_value(point.name)

        # 寫入 raw registers
        self._block.set_raw(address, values)

        # 觸發 on_write 回調
        for point in affected:
            new_value = self._block.get_value(point.name)
            old_value = old_values[point.name]
            if old_value != new_value:
                self._simulator.on_write(point.name, old_value, new_value)


class SimulationServer(AsyncLifecycleMixin):
    """
    Modbus TCP 模擬伺服器

    封裝 pymodbus TCP server，提供：
    - 多設備模擬器管理（per unit_id）
    - 定期 tick loop 呼叫所有 simulator 的 update()
    - 支援 MicrogridSimulator 聯動模式
    - AsyncLifecycleMixin 生命週期管理
    """

    def __init__(self, config: ServerConfig | None = None) -> None:
        self._config = config or ServerConfig()
        self._simulators: dict[int, BaseDeviceSimulator] = {}
        self._microgrid: MicrogridSimulator | None = None
        self._server: Any = None
        self._tick_task: asyncio.Task[None] | None = None
        self._running = False

    @property
    def config(self) -> ServerConfig:
        return self._config

    @property
    def simulators(self) -> dict[int, BaseDeviceSimulator]:
        return dict(self._simulators)

    @property
    def is_running(self) -> bool:
        return self._running

    def add_simulator(self, simulator: BaseDeviceSimulator) -> None:
        """註冊設備模擬器"""
        if simulator.unit_id in self._simulators:
            raise ValueError(f"Unit ID {simulator.unit_id} already registered")
        self._simulators[simulator.unit_id] = simulator

    def set_microgrid(self, microgrid: MicrogridSimulator) -> None:
        """
        設定 MicrogridSimulator 聯動模式

        自動註冊所有 microgrid 中的 simulators。
        Tick loop 將呼叫 microgrid.update() 而非個別 simulator.update()。
        """
        self._microgrid = microgrid
        for sim in microgrid.all_simulators:
            if sim.unit_id not in self._simulators:
                self._simulators[sim.unit_id] = sim

    async def _on_start(self) -> None:
        """啟動 server"""
        _ensure_pymodbus_imported()

        # 建立 per-simulator datablock 和 context
        slaves = {}
        for unit_id, sim in self._simulators.items():
            data_block = _create_datablock(sim)
            # ModbusDeviceContext: di=discrete inputs, co=coils, hr=holding registers, ir=input registers
            slave_ctx = get_ModbusDeviceContext()(  # type: ignore[misc]
                di=data_block,
                co=data_block,
                hr=data_block,
                ir=data_block,
            )
            slaves[unit_id] = slave_ctx

        if not slaves:
            logger.warning("No simulators registered, server will start with empty context")
            data_block = _create_empty_pymodbus_block()
            slaves[0] = get_ModbusDeviceContext()(di=data_block, co=data_block, hr=data_block, ir=data_block)  # type: ignore[misc]

        server_ctx = get_ModbusServerContext()(devices=slaves, single=False)  # type: ignore[misc]

        self._server = get_ModbusTcpServer()(  # type: ignore[misc]
            context=server_ctx,
            address=(self._config.host, self._config.port),
        )

        # 啟動 server（背景執行）
        await self._server.serve_forever(background=True)
        self._running = True

        # 啟動 tick loop
        self._tick_task = asyncio.create_task(self._tick_loop())

        logger.info(f"SimulationServer started on {self._config.host}:{self._config.port}")

    async def _on_stop(self) -> None:
        """停止 server"""
        self._running = False

        # 取消 tick loop
        if self._tick_task is not None:
            self._tick_task.cancel()
            try:
                await self._tick_task
            except asyncio.CancelledError:
                pass
            self._tick_task = None

        # 關閉 server
        if self._server is not None:
            await self._server.shutdown()
            self._server = None

        logger.info("SimulationServer stopped")

    async def _tick_loop(self) -> None:
        """定期 tick loop"""
        interval = self._config.tick_interval
        try:
            while self._running:
                await self._tick(interval)
                await asyncio.sleep(interval)
        except asyncio.CancelledError:
            pass

    async def _tick(self, interval: float) -> None:
        """執行一次 tick"""
        if self._microgrid is not None:
            await self._microgrid.update(interval)
        else:
            for sim in self._simulators.values():
                await sim.update()

    async def serve(self) -> None:
        """持續運行直到被停止"""
        try:
            while self._running:
                await asyncio.sleep(1.0)
        except asyncio.CancelledError:
            pass


def _create_datablock(simulator: BaseDeviceSimulator) -> Any:
    """建立繼承 BaseModbusDataBlock 的 SimulatorDataBlock

    注意: pymodbus 3.12 的 ModbusDeviceContext 在呼叫 datablock 前會
    自動 address += 1，因此 datablock 內需要 address -= 1 來補償。
    """
    wrapper = SimulatorDataBlock(simulator)

    class _PymodbusDataBlock(get_BaseModbusDataBlock()):  # type: ignore[misc]
        def __init__(self):
            self.address = 0
            self.default_value = 0
            self.values = wrapper._block.registers

        def reset(self):
            wrapper.reset()

        def getValues(self, address, count=1):
            return wrapper.getValues(address - 1, count)

        def setValues(self, address, values):
            return wrapper.setValues(address - 1, values)

    return _PymodbusDataBlock()


def _create_empty_pymodbus_block() -> Any:
    """建立空的 pymodbus DataBlock"""

    class _EmptyBlock(get_BaseModbusDataBlock()):  # type: ignore[misc]
        def __init__(self):
            self.address = 0
            self.default_value = 0
            self.values = [0] * 100

        def reset(self):
            pass

        def getValues(self, address, count=1):
            return [0] * count

        def setValues(self, address, values):
            pass

    return _EmptyBlock()


__all__ = ["SimulationServer", "SimulatorDataBlock"]



# =============== Modbus Server - Behaviors ===============
#
# 可組合的模擬行為

from .alarm import AlarmBehavior
from .curve import CurveBehavior
from .noise import NoiseBehavior
from .ramp import RampBehavior

__all__ = [
    "AlarmBehavior",
    "CurveBehavior",
    "NoiseBehavior",
    "RampBehavior",
]



# =============== CAN Client - Base ===============
#
# CAN 客戶端抽象基底類別

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable

from csp_lib.can.config import CANFrame


class AsyncCANClientBase(ABC):
    """
    CAN 客戶端抽象基底

    定義連線管理、被動監聽、主動發送、請求-回應四組介面。
    """

    # =============== 連線管理 ===============

    @abstractmethod
    async def connect(self) -> None:
        """建立連線"""

    @abstractmethod
    async def disconnect(self) -> None:
        """斷開連線"""

    @abstractmethod
    async def is_connected(self) -> bool:
        """檢查是否已連線"""

    # =============== 被動監聽 ===============

    @abstractmethod
    async def start_listener(self) -> None:
        """啟動背景接收"""

    @abstractmethod
    async def stop_listener(self) -> None:
        """停止背景接收"""

    @abstractmethod
    def subscribe(self, can_id: int, handler: Callable[[CANFrame], Any]) -> Callable[[], None]:
        """
        訂閱指定 CAN ID 的訊框

        Args:
            can_id: 要訂閱的 CAN ID
            handler: 收到訊框時的回調函數

        Returns:
            取消訂閱的函數
        """

    # =============== 主動發送 ===============

    @abstractmethod
    async def send(self, can_id: int, data: bytes) -> None:
        """
        發送 CAN 訊框

        Args:
            can_id: CAN 訊框 ID
            data: 訊框資料（最多 8 bytes）
        """

    # =============== 請求-回應 ===============

    @abstractmethod
    async def request(
        self,
        can_id: int,
        data: bytes,
        response_id: int,
        timeout: float = 1.0,
    ) -> CANFrame:
        """
        發送請求並等待回應

        Args:
            can_id: 發送的 CAN ID
            data: 請求資料
            response_id: 期望回應的 CAN ID
            timeout: 逾時時間（秒）

        Returns:
            回應訊框

        Raises:
            CANTimeoutError: 等待回應逾時
        """


__all__ = [
    "AsyncCANClientBase",
]



"""
CommandProcessor — Post-Protection 命令處理器 Protocol

定義在 ProtectionGuard 和 CommandRouter 之間的處理管線。
典型用途：功率補償、命令日誌、審計追蹤等。

執行流程::

    Strategy.execute()
      → ProtectionGuard.apply()
      → [CommandProcessor 1] → [CommandProcessor 2] → ...
      → CommandRouter.route()

Usage::

    class MyCompensator:
        async def process(self, command, context):
            compensated_p = self._compensate(command.p_target)
            return command.with_p(compensated_p)

    config = SystemControllerConfig(
        post_protection_processors=[MyCompensator()],
    )
"""

from __future__ import annotations

from typing import Protocol, runtime_checkable

from .command import Command
from .context import StrategyContext


@runtime_checkable
class CommandProcessor(Protocol):
    """
    Post-Protection 命令處理器

    在 ProtectionGuard 和 CommandRouter 之間對命令做額外處理。
    實作此 Protocol 的物件可在 SystemControllerConfig 中註冊。

    方法語義：
    - 輸入：經保護鏈處理後的 Command 和當前 StrategyContext
    - 輸出：處理後的 Command（可修改 p_target / q_target）
    - 拋出例外時：SystemController 會 log 並跳過此 processor，繼續後續處理
    """

    async def process(self, command: Command, context: StrategyContext) -> Command:
        """
        處理命令

        Args:
            command: 經保護鏈處理後的命令
            context: 當前策略上下文

        Returns:
            處理後的命令
        """
        ...


__all__ = ["CommandProcessor"]



# =============== Manager Alarm - Persistence ===============
#
# 告警持久化管理器
#
# 提供告警事件的自動持久化功能：
#   - AlarmPersistenceManager: 訂閱設備事件並自動寫入/更新告警記錄
#
# 設計模式：
#   - 觀察者模式：訂閱 AsyncModbusDevice 的連線/告警事件
#   - 事件驅動：斷線/告警觸發 → 寫入 DB，恢復/解除 → 更新 resolved_timestamp
from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Callable

from csp_lib.core import get_logger
from csp_lib.equipment.alarm import AlarmLevel
from csp_lib.equipment.device import AsyncModbusDevice
from csp_lib.equipment.device.events import (
    EVENT_ALARM_CLEARED,
    EVENT_ALARM_TRIGGERED,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    ConnectedPayload,
    DeviceAlarmPayload,
    DisconnectPayload,
)
from csp_lib.manager.base import DeviceEventSubscriber

from .config import AlarmPersistenceConfig
from .repository import AlarmRepository
from .schema import AlarmRecord, AlarmType

if TYPE_CHECKING:
    from csp_lib.notification import NotificationSender

logger = get_logger(__name__)


class _NotifyEvent(str, Enum):
    """內部通知事件類型（避免頂層 import 循環）"""

    TRIGGERED = "triggered"
    RESOLVED = "resolved"


class AlarmPersistenceManager(DeviceEventSubscriber):
    """
    告警持久化管理器

    自動將設備事件持久化至資料庫。採用觀察者模式訂閱 AsyncModbusDevice
    的連線與告警事件，實現事件驅動的告警管理。

    職責：
        1. 訂閱多個 AsyncModbusDevice 的事件
        2. 斷線/告警觸發 → 寫入 DB（新增告警記錄）
        3. 恢復/告警解除 → 更新 resolved_timestamp（解除告警）

    Attributes:
        DISCONNECT_CODE: 斷線告警的固定代碼
        DISCONNECT_NAME: 斷線告警的顯示名稱
    """

    def __init__(
        self,
        repository: AlarmRepository,
        dispatcher: NotificationSender | None = None,
        config: AlarmPersistenceConfig | None = None,
    ) -> None:
        """
        初始化告警持久化管理器

        Args:
            repository: 告警資料存取層（遵循 AlarmRepository Protocol）
            dispatcher: 通知分發器（可選），用於告警觸發/解除時發送通知
            config: 告警持久化配置（可選，預設使用 AlarmPersistenceConfig()）
        """
        super().__init__()
        self._repository = repository
        self._dispatcher = dispatcher
        self._config = config or AlarmPersistenceConfig()

    # ================ 訂閱管理 ================

    def _register_events(self, device: AsyncModbusDevice) -> list[Callable[[], None]]:
        """註冊設備的連線/斷線與告警觸發/解除事件"""
        logger.info(f"告警持久化管理器已訂閱設備: {device.device_id}")
        return [
            device.on(EVENT_DISCONNECTED, self._on_disconnected),
            device.on(EVENT_CONNECTED, self._on_connected),
            device.on(EVENT_ALARM_TRIGGERED, self._on_alarm_triggered),
            device.on(EVENT_ALARM_CLEARED, self._on_alarm_cleared),
        ]

    def _on_unsubscribe(self, device_id: str) -> None:
        logger.info(f"告警持久化管理器已取消訂閱設備: {device_id}")

    # ================ 事件處理器 ================

    async def _on_disconnected(self, payload: DisconnectPayload) -> None:
        """
        處理設備斷線事件

        建立斷線類型的告警記錄並寫入資料庫。

        Args:
            payload: 斷線事件資料（包含 device_id、reason、timestamp）
        """
        record = AlarmRecord(
            alarm_key=AlarmRecord.make_key(payload.device_id, AlarmType.DISCONNECT, self._config.disconnect_code),
            device_id=payload.device_id,
            alarm_type=AlarmType.DISCONNECT,
            alarm_code=self._config.disconnect_code,
            name=self._config.disconnect_name,
            level=AlarmLevel.WARNING,
            description=payload.reason,
            timestamp=payload.timestamp,
        )
        await self._create_alarm(record)

    async def _on_connected(self, payload: ConnectedPayload) -> None:
        """
        處理設備連線事件

        解除對應的斷線告警。

        Args:
            payload: 連線事件資料（包含 device_id、timestamp）
        """
        key = AlarmRecord.make_key(payload.device_id, AlarmType.DISCONNECT, self._config.disconnect_code)
        await self._resolve_alarm(key, payload.timestamp)

    async def _on_alarm_triggered(self, payload: DeviceAlarmPayload) -> None:
        """
        處理設備告警觸發事件

        建立設備內部告警記錄並寫入資料庫。

        Args:
            payload: 告警事件資料（包含 device_id、alarm_event、timestamp）
        """
        alarm = payload.alarm_event.alarm
        record = AlarmRecord(
            alarm_key=AlarmRecord.make_key(payload.device_id, AlarmType.DEVICE_ALARM, alarm.code),
            device_id=payload.device_id,
            alarm_type=AlarmType.DEVICE_ALARM,
            alarm_code=alarm.code,
            name=alarm.name,
            level=alarm.level,
            description=alarm.description,
            timestamp=payload.timestamp,
        )
        await self._create_alarm(record)

    async def _on_alarm_cleared(self, payload: DeviceAlarmPayload) -> None:
        """
        處理設備告警解除事件

        解除對應的設備內部告警。

        Args:
            payload: 告警事件資料（包含 device_id、alarm_event、timestamp）
        """
        alarm = payload.alarm_event.alarm
        key = AlarmRecord.make_key(payload.device_id, AlarmType.DEVICE_ALARM, alarm.code)
        await self._resolve_alarm(key, payload.timestamp)

    # ================ 私有方法 ================

    async def _create_alarm(self, record: AlarmRecord) -> None:
        """
        建立告警記錄

        透過 repository 寫入告警記錄。若為新告警則記錄 log 並發送通知。

        Args:
            record: 告警記錄
        """
        _, is_new = await self._repository.upsert(record)
        if is_new:
            logger.info(f"告警持久化管理器已新增告警: {record.alarm_key}")
            await self._notify(record, _NotifyEvent.TRIGGERED)

    async def _resolve_alarm(self, alarm_key: str, resolved_at: datetime) -> None:
        """
        解除告警記錄

        透過 repository 更新告警狀態為已解除。若成功則記錄 log 並發送通知。

        Args:
            alarm_key: 告警唯一鍵
            resolved_at: 解除時間
        """
        success = await self._repository.resolve(alarm_key, resolved_at)
        if success:
            logger.info(f"告警持久化管理器已解除告警: {alarm_key}")
            await self._notify_resolved(alarm_key, resolved_at)

    async def _notify(self, record: AlarmRecord, event: _NotifyEvent) -> None:
        """發送告警通知（非阻塞，失敗僅記 log）"""
        if self._dispatcher is None:
            return
        try:
            from csp_lib.notification import NotificationDispatcher, NotificationEvent

            notification = NotificationDispatcher.from_alarm_record(record, NotificationEvent(event.value))
            await self._dispatcher.dispatch(notification)
        except Exception:
            logger.opt(exception=True).warning(f"告警通知發送失敗: {record.alarm_key}")

    async def _notify_resolved(self, alarm_key: str, resolved_at: datetime) -> None:
        """發送解除通知（從 alarm_key 拆解 device_id，不查 DB）"""
        if self._dispatcher is None:
            return
        try:
            from csp_lib.notification import Notification, NotificationEvent

            # alarm_key 格式: "<device_id>:<alarm_type>:<alarm_code>"
            parts = alarm_key.split(":")
            device_id = parts[0] if parts else alarm_key

            notification = Notification(
                title=f"[RESOLVED] {device_id} 告警解除",
                body=f"告警 {alarm_key} 已解除",
                level=AlarmLevel.INFO,
                device_id=device_id,
                alarm_key=alarm_key,
                event=NotificationEvent.RESOLVED,
                occurred_at=resolved_at,
            )
            await self._dispatcher.dispatch(notification)
        except Exception:
            logger.opt(exception=True).warning(f"告警解除通知發送失敗: {alarm_key}")



# =============== Frozen Dataclass Config Tests ===============
#
# Comprehensive tests for 6 frozen dataclass configs with __post_init__ validation:
#   1. PointGrouperConfig
#   2. AlarmPersistenceConfig
#   3. CommandAdapterConfig
#   4. StateSyncConfig
#   5. ScheduleServiceConfig
#   6. NotificationConfig

from __future__ import annotations

from dataclasses import FrozenInstanceError

import pytest

from csp_lib.equipment.transport.config import PointGrouperConfig
from csp_lib.manager.alarm.config import AlarmPersistenceConfig
from csp_lib.manager.command.adapters.config import CommandAdapterConfig
from csp_lib.manager.schedule.config import ScheduleServiceConfig
from csp_lib.manager.state.config import StateSyncConfig
from csp_lib.notification.config import NotificationConfig

# ---------------------------------------------------------------------------
# 1. PointGrouperConfig
# ---------------------------------------------------------------------------


class TestPointGrouperConfig:
    """PointGrouperConfig 單元測試"""

    def test_default_values(self):
        """預設 fc_max_length 應包含 4 個功能碼，且值正確"""
        config = PointGrouperConfig()
        assert config.fc_max_length == {
            1: 2000,
            2: 2000,
            3: 125,
            4: 125,
        }

    def test_default_dict_is_independent_copy(self):
        """每次建立的預設 dict 應為獨立副本"""
        c1 = PointGrouperConfig()
        c2 = PointGrouperConfig()
        assert c1.fc_max_length is not c2.fc_max_length

    def test_custom_fc_max_length(self):
        """自訂 fc_max_length 應被接受"""
        custom = {3: 50, 4: 80}
        config = PointGrouperConfig(fc_max_length=custom)
        assert config.fc_max_length == {3: 50, 4: 80}

    def test_single_entry(self):
        """僅一個功能碼也應有效"""
        config = PointGrouperConfig(fc_max_length={1: 1})
        assert config.fc_max_length == {1: 1}

    def test_empty_dict_is_valid(self):
        """空字典不違反驗證（無元素可檢查）"""
        config = PointGrouperConfig(fc_max_length={})
        assert config.fc_max_length == {}

    @pytest.mark.parametrize("bad_length", [0, -1, -100])
    def test_zero_or_negative_length_raises(self, bad_length: int):
        """fc_max_length 中任何值 <= 0 應拋 ValueError"""
        with pytest.raises(ValueError, match="最大讀取長度必須大於 0"):
            PointGrouperConfig(fc_max_length={3: bad_length})

    def test_multiple_entries_one_invalid(self):
        """多筆中只要有一筆 <= 0 就應拋 ValueError"""
        with pytest.raises(ValueError, match="最大讀取長度必須大於 0"):
            PointGrouperConfig(fc_max_length={1: 2000, 3: 0})

    def test_error_message_contains_fc_code(self):
        """錯誤訊息應包含功能碼編號"""
        with pytest.raises(ValueError, match="功能碼 99"):
            PointGrouperConfig(fc_max_length={99: -5})

    def test_frozen_immutability(self):
        """PointGrouperConfig 應為不可變"""
        config = PointGrouperConfig()
        with pytest.raises(FrozenInstanceError):
            config.fc_max_length = {}  # type: ignore[misc]


# ---------------------------------------------------------------------------
# 2. AlarmPersistenceConfig
# ---------------------------------------------------------------------------


class TestAlarmPersistenceConfig:
    """AlarmPersistenceConfig 單元測試"""

    def test_default_values(self):
        """預設值應正確"""
        config = AlarmPersistenceConfig()
        assert config.disconnect_code == "DISCONNECT"
        assert config.disconnect_name == "設備斷線"

    def test_custom_values(self):
        """自訂值應被接受"""
        config = AlarmPersistenceConfig(disconnect_code="DC_001", disconnect_name="Link Down")
        assert config.disconnect_code == "DC_001"
        assert config.disconnect_name == "Link Down"

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_disconnect_code_raises(self, empty_val):
        """空的 disconnect_code 應拋 ValueError"""
        with pytest.raises(ValueError, match="disconnect_code 不可為空"):
            AlarmPersistenceConfig(disconnect_code=empty_val)  # type: ignore[arg-type]

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_disconnect_name_raises(self, empty_val):
        """空的 disconnect_name 應拋 ValueError"""
        with pytest.raises(ValueError, match="disconnect_name 不可為空"):
            AlarmPersistenceConfig(disconnect_name=empty_val)  # type: ignore[arg-type]

    def test_whitespace_only_disconnect_code_accepted(self):
        """僅含空白的 disconnect_code 視為非空（truthy），應被接受"""
        config = AlarmPersistenceConfig(disconnect_code="  ")
        assert config.disconnect_code == "  "

    def test_frozen_disconnect_code(self):
        """不可修改 disconnect_code"""
        config = AlarmPersistenceConfig()
        with pytest.raises(FrozenInstanceError):
            config.disconnect_code = "NEW"  # type: ignore[misc]

    def test_frozen_disconnect_name(self):
        """不可修改 disconnect_name"""
        config = AlarmPersistenceConfig()
        with pytest.raises(FrozenInstanceError):
            config.disconnect_name = "NEW"  # type: ignore[misc]


# ---------------------------------------------------------------------------
# 3. CommandAdapterConfig
# ---------------------------------------------------------------------------


class TestCommandAdapterConfig:
    """CommandAdapterConfig 單元測試"""

    def test_default_values(self):
        """預設值應正確"""
        config = CommandAdapterConfig()
        assert config.command_channel == "channel:commands:write"
        assert config.result_channel == "channel:commands:result"

    def test_custom_values(self):
        """自訂 channel 名稱應被接受"""
        config = CommandAdapterConfig(command_channel="cmd:in", result_channel="cmd:out")
        assert config.command_channel == "cmd:in"
        assert config.result_channel == "cmd:out"

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_command_channel_raises(self, empty_val):
        """空的 command_channel 應拋 ValueError"""
        with pytest.raises(ValueError, match="command_channel 不可為空"):
            CommandAdapterConfig(command_channel=empty_val)  # type: ignore[arg-type]

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_result_channel_raises(self, empty_val):
        """空的 result_channel 應拋 ValueError"""
        with pytest.raises(ValueError, match="result_channel 不可為空"):
            CommandAdapterConfig(result_channel=empty_val)  # type: ignore[arg-type]

    def test_both_empty_raises(self):
        """兩個 channel 同時為空，應先報 command_channel"""
        with pytest.raises(ValueError, match="command_channel 不可為空"):
            CommandAdapterConfig(command_channel="", result_channel="")

    def test_frozen_command_channel(self):
        """不可修改 command_channel"""
        config = CommandAdapterConfig()
        with pytest.raises(FrozenInstanceError):
            config.command_channel = "new"  # type: ignore[misc]

    def test_frozen_result_channel(self):
        """不可修改 result_channel"""
        config = CommandAdapterConfig()
        with pytest.raises(FrozenInstanceError):
            config.result_channel = "new"  # type: ignore[misc]


# ---------------------------------------------------------------------------
# 4. StateSyncConfig
# ---------------------------------------------------------------------------


class TestStateSyncConfig:
    """StateSyncConfig 單元測試"""

    def test_default_values(self):
        """預設值應正確"""
        config = StateSyncConfig()
        assert config.state_ttl == 60
        assert config.online_ttl == 60

    def test_custom_values(self):
        """自訂值應被接受"""
        config = StateSyncConfig(state_ttl=120, online_ttl=30)
        assert config.state_ttl == 120
        assert config.online_ttl == 30

    def test_boundary_value_one(self):
        """最小合法值 1 應被接受"""
        config = StateSyncConfig(state_ttl=1, online_ttl=1)
        assert config.state_ttl == 1
        assert config.online_ttl == 1

    def test_large_ttl(self):
        """大 TTL 值應被接受"""
        config = StateSyncConfig(state_ttl=86400, online_ttl=86400)
        assert config.state_ttl == 86400

    @pytest.mark.parametrize("bad_ttl", [0, -1, -100])
    def test_zero_or_negative_state_ttl_raises(self, bad_ttl: int):
        """state_ttl <= 0 應拋 ValueError"""
        with pytest.raises(ValueError, match="state_ttl 必須大於 0"):
            StateSyncConfig(state_ttl=bad_ttl)

    @pytest.mark.parametrize("bad_ttl", [0, -1, -100])
    def test_zero_or_negative_online_ttl_raises(self, bad_ttl: int):
        """online_ttl <= 0 應拋 ValueError"""
        with pytest.raises(ValueError, match="online_ttl 必須大於 0"):
            StateSyncConfig(online_ttl=bad_ttl)

    def test_error_message_contains_value(self):
        """錯誤訊息應包含實際收到的值"""
        with pytest.raises(ValueError, match="-42"):
            StateSyncConfig(state_ttl=-42)

    def test_both_invalid_raises_state_ttl_first(self):
        """兩者同時不合法，應先拋 state_ttl 的錯誤"""
        with pytest.raises(ValueError, match="state_ttl"):
            StateSyncConfig(state_ttl=0, online_ttl=0)

    def test_frozen_state_ttl(self):
        """不可修改 state_ttl"""
        config = StateSyncConfig()
        with pytest.raises(FrozenInstanceError):
            config.state_ttl = 999  # type: ignore[misc]

    def test_frozen_online_ttl(self):
        """不可修改 online_ttl"""
        config = StateSyncConfig()
        with pytest.raises(FrozenInstanceError):
            config.online_ttl = 999  # type: ignore[misc]


# ---------------------------------------------------------------------------
# 5. ScheduleServiceConfig
# ---------------------------------------------------------------------------


class TestScheduleServiceConfig:
    """ScheduleServiceConfig 單元測試"""

    def test_default_poll_interval_and_timezone(self):
        """預設 poll_interval 與 timezone_name 應正確（site_id 必須提供）"""
        config = ScheduleServiceConfig(site_id="site_001")
        assert config.poll_interval == 30.0
        assert config.timezone_name == "Asia/Taipei"

    def test_default_site_id_is_empty_and_raises(self):
        """預設 site_id 為空字串，建立時應拋 ValueError"""
        with pytest.raises(ValueError, match="site_id 不可為空"):
            ScheduleServiceConfig()

    def test_custom_values(self):
        """自訂值應被接受"""
        config = ScheduleServiceConfig(
            site_id="plant_A",
            poll_interval=10.0,
            timezone_name="UTC",
        )
        assert config.site_id == "plant_A"
        assert config.poll_interval == 10.0
        assert config.timezone_name == "UTC"

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_site_id_raises(self, empty_val):
        """空 site_id 應拋 ValueError"""
        with pytest.raises(ValueError, match="site_id 不可為空"):
            ScheduleServiceConfig(site_id=empty_val)  # type: ignore[arg-type]

    @pytest.mark.parametrize("bad_interval", [0, 0.0, -1, -0.5, -100.0])
    def test_zero_or_negative_poll_interval_raises(self, bad_interval: float):
        """poll_interval <= 0 應拋 ValueError"""
        with pytest.raises(ValueError, match="poll_interval 必須大於 0"):
            ScheduleServiceConfig(site_id="s1", poll_interval=bad_interval)

    def test_boundary_poll_interval_just_above_zero(self):
        """極小正數 poll_interval 應被接受"""
        config = ScheduleServiceConfig(site_id="s1", poll_interval=0.001)
        assert config.poll_interval == pytest.approx(0.001)

    def test_error_message_contains_poll_interval_value(self):
        """錯誤訊息應包含實際值"""
        with pytest.raises(ValueError, match="-5"):
            ScheduleServiceConfig(site_id="s1", poll_interval=-5)

    def test_frozen_site_id(self):
        """不可修改 site_id"""
        config = ScheduleServiceConfig(site_id="s1")
        with pytest.raises(FrozenInstanceError):
            config.site_id = "s2"  # type: ignore[misc]

    def test_frozen_poll_interval(self):
        """不可修改 poll_interval"""
        config = ScheduleServiceConfig(site_id="s1")
        with pytest.raises(FrozenInstanceError):
            config.poll_interval = 99.0  # type: ignore[misc]

    def test_frozen_timezone_name(self):
        """不可修改 timezone_name"""
        config = ScheduleServiceConfig(site_id="s1")
        with pytest.raises(FrozenInstanceError):
            config.timezone_name = "UTC"  # type: ignore[misc]


# ---------------------------------------------------------------------------
# 6. NotificationConfig
# ---------------------------------------------------------------------------


class TestNotificationConfig:
    """NotificationConfig 單元測試"""

    def test_default_values(self):
        """預設值應正確"""
        config = NotificationConfig()
        assert config.triggered_label == "觸發"
        assert config.resolved_label == "解除"
        assert config.title_template == "[{level}] {device_id} {name} - {event_label}"

    def test_custom_values(self):
        """自訂值應被接受"""
        config = NotificationConfig(
            triggered_label="TRIGGERED",
            resolved_label="RESOLVED",
            title_template="{device_id}: {name}",
        )
        assert config.triggered_label == "TRIGGERED"
        assert config.resolved_label == "RESOLVED"
        assert config.title_template == "{device_id}: {name}"

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_triggered_label_raises(self, empty_val):
        """空 triggered_label 應拋 ValueError"""
        with pytest.raises(ValueError, match="triggered_label 不可為空"):
            NotificationConfig(triggered_label=empty_val)  # type: ignore[arg-type]

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_resolved_label_raises(self, empty_val):
        """空 resolved_label 應拋 ValueError"""
        with pytest.raises(ValueError, match="resolved_label 不可為空"):
            NotificationConfig(resolved_label=empty_val)  # type: ignore[arg-type]

    @pytest.mark.parametrize("empty_val", ["", None])
    def test_empty_title_template_raises(self, empty_val):
        """空 title_template 應拋 ValueError"""
        with pytest.raises(ValueError, match="title_template 不可為空"):
            NotificationConfig(title_template=empty_val)  # type: ignore[arg-type]

    def test_all_empty_raises_triggered_first(self):
        """三個欄位同時為空，應先報 triggered_label 的錯誤"""
        with pytest.raises(ValueError, match="triggered_label 不可為空"):
            NotificationConfig(triggered_label="", resolved_label="", title_template="")

    def test_whitespace_only_is_accepted(self):
        """僅含空白的字串視為非空（truthy），應被接受"""
        config = NotificationConfig(triggered_label=" ", resolved_label=" ", title_template=" ")
        assert config.triggered_label == " "

    def test_frozen_triggered_label(self):
        """不可修改 triggered_label"""
        config = NotificationConfig()
        with pytest.raises(FrozenInstanceError):
            config.triggered_label = "X"  # type: ignore[misc]

    def test_frozen_resolved_label(self):
        """不可修改 resolved_label"""
        config = NotificationConfig()
        with pytest.raises(FrozenInstanceError):
            config.resolved_label = "X"  # type: ignore[misc]

    def test_frozen_title_template(self):
        """不可修改 title_template"""
        config = NotificationConfig()
        with pytest.raises(FrozenInstanceError):
            config.title_template = "X"  # type: ignore[misc]



# =============== Equipment Device - Base ===============
#
# 非同步 Modbus 設備
#
# 整合讀寫、告警、事件的完整設備抽象

from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Sequence

from csp_lib.core import get_logger
from csp_lib.core.errors import CommunicationError, ConfigurationError, DeviceConnectionError
from csp_lib.core.health import HealthReport, HealthStatus
from csp_lib.equipment.alarm import AlarmEvaluator, AlarmStateManager
from csp_lib.equipment.processing import AggregatorPipeline
from csp_lib.equipment.transport import (
    GroupReader,
    PointGrouper,
    ReadScheduler,
    ValidatedWriter,
)

from .capability import Capability, CapabilityBinding
from .config import DeviceConfig
from .events import (
    EVENT_CAPABILITY_ADDED,
    EVENT_CAPABILITY_REMOVED,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_POINT_TOGGLED,
    EVENT_READ_COMPLETE,
    EVENT_READ_ERROR,
    EVENT_RECONFIGURED,
    EVENT_RESTARTED,
    EVENT_VALUE_CHANGE,
    AsyncHandler,
    CapabilityChangedPayload,
    ConnectedPayload,
    DeviceEventEmitter,
    DisconnectPayload,
    PointToggledPayload,
    ReadCompletePayload,
    ReadErrorPayload,
    ReconfiguredPayload,
    RestartedPayload,
    ValueChangePayload,
)
from .mixins import AlarmMixin, WriteMixin

if TYPE_CHECKING:
    from csp_lib.equipment.core import PointMetadata, ReadPoint, WritePoint
    from csp_lib.modbus.clients.base import AsyncModbusClientBase
    from csp_lib.modbus.types import ModbusDataType

    from .action import DOActionConfig

logger = get_logger(__name__)


@dataclass(frozen=True, slots=True)
class PointInfo:
    """點位詳細資訊"""

    name: str
    address: int
    data_type: ModbusDataType
    direction: str  # "read" | "write" | "read_write"
    enabled: bool
    read_group: str
    metadata: PointMetadata | None


@dataclass(frozen=True, slots=True)
class ReconfigureSpec:
    """重新配置規格"""

    always_points: Sequence[ReadPoint] | None = None
    rotating_points: Sequence[Sequence[ReadPoint]] | None = None
    write_points: Sequence[WritePoint] | None = None
    alarm_evaluators: Sequence[AlarmEvaluator] | None = None
    capability_bindings: Sequence[CapabilityBinding] | None = None


class AsyncModbusDevice(AlarmMixin, WriteMixin):
    """
    非同步 Modbus 設備

    整合讀寫、告警、事件的完整設備抽象。

    特點：
        - 定期讀取循環
        - 告警狀態管理（含遲滯）
        - 事件驅動通知
        - 動態點位排程
        - 設備寫入管理
        - 高階 Action 指令支援

    Class Attributes:
        ACTIONS: 動作名稱對應方法名稱的映射，子類別可覆寫。
            例如: {"start": "set_generator_on", "stop": "set_generator_off"}
    """

    # 子類別可覆寫，定義支援的 action -> method 映射
    ACTIONS: dict[str, str] = {}

    def __init__(
        self,
        config: DeviceConfig,
        client: AsyncModbusClientBase,
        always_points: Sequence[ReadPoint] = (),
        rotating_points: Sequence[Sequence[ReadPoint]] = (),
        write_points: Sequence[WritePoint] = (),
        alarm_evaluators: Sequence[AlarmEvaluator] = (),
        aggregator_pipeline: AggregatorPipeline | None = None,
        capability_bindings: Sequence[CapabilityBinding] = (),
    ):
        self._config = config
        self._client = client

        # 儲存原始點位定義
        self._read_points_always: tuple[ReadPoint, ...] = tuple(always_points)
        self._read_points_rotating: tuple[tuple[ReadPoint, ...], ...] = tuple(tuple(pts) for pts in rotating_points)

        # 建立排程器（自動分組）
        self._grouper = PointGrouper()
        self._scheduler = ReadScheduler(
            always_groups=self._grouper.group(list(always_points)),
            rotating_groups=[self._grouper.group(list(points)) for points in rotating_points],
        )

        # 通訊用元件
        self._reader = GroupReader(
            client=client,
            unit_id=config.unit_id,
            address_offset=config.address_offset,
        )
        self._writer = ValidatedWriter(
            client=client,
            unit_id=config.unit_id,
            address_offset=config.address_offset,
        )

        # 寫入點位查詢表
        self._write_points = {write_point.name: write_point for write_point in write_points}

        # 告警管理
        self._alarm_manager = AlarmStateManager()
        self._alarm_evaluators = list(alarm_evaluators)
        for evaluator in self._alarm_evaluators:
            self._alarm_manager.register_alarms(evaluator.get_alarms())

        # 設備層級聚合處理
        self._aggregator_pipeline = aggregator_pipeline

        # 能力綁定
        self._capability_bindings: dict[str, CapabilityBinding] = {b.capability.name: b for b in capability_bindings}

        # 點位開關
        self._disabled_points: set[str] = set()

        # 事件
        self._emitter = DeviceEventEmitter()

        # 設備狀態
        self._latest_values: dict[str, Any] = {}
        self._client_connected = False  # Socket 層級：client.connect() 是否成功
        self._device_responsive = False  # 通訊層級：設備是否有回應
        self._consecutive_failures = 0  # 連續讀取失敗次數
        self._last_failure_time: float | None = None  # 最後一次讀取失敗的時間（monotonic）
        self._status_lock = asyncio.Lock()  # 保護狀態欄位的非同步鎖
        self._stop_event = asyncio.Event()
        self._read_task: asyncio.Task[None] | None = None

        # DO 動作（WriteMixin）
        self._do_actions: dict[str, DOActionConfig] = {}
        self._pulse_tasks: list[asyncio.Task[None]] = []

    # =============== Properties ===============

    @property
    def device_id(self) -> str:
        """設備唯一識別碼(DB/通訊用)"""
        return self._config.device_id

    @property
    def is_connected(self) -> bool:
        """Client 是否已連線（Socket 層級）"""
        return self._client_connected

    @property
    def is_responsive(self) -> bool:
        """設備是否有回應（通訊層級）"""
        return self._device_responsive

    @property
    def is_disconnected(self) -> bool:
        """Client 是否已斷線"""
        return not self._client_connected

    @property
    def is_unreachable(self) -> bool:
        """設備是否無回應（連續失敗達閾值）"""
        return self._client_connected and not self._device_responsive

    @property
    def is_healthy(self) -> bool:
        """設備是否健康"""
        return self.is_connected and self.is_responsive and not self.is_protected

    @property
    def latest_values(self) -> dict[str, Any]:
        """最新讀取值"""
        return self._latest_values.copy()

    @property
    def is_running(self) -> bool:
        """讀取循環是否運行中"""
        return self._read_task is not None and not self._read_task.done()

    @property
    def should_attempt_read(self) -> bool:
        """
        是否應嘗試讀取（用於群組讀取排程）

        回應中的設備或尚未失敗的設備永遠回傳 True。
        無回應的設備在超過 reconnect_interval 後才回傳 True，避免 timeout 阻塞群組讀取。
        """
        if self._device_responsive:
            return True
        if self._last_failure_time is None:
            return True
        return (time.monotonic() - self._last_failure_time) >= self._config.reconnect_interval

    # =============== Capabilities ===============

    @property
    def capabilities(self) -> dict[str, CapabilityBinding]:
        """所有已綁定的能力"""
        return dict(self._capability_bindings)

    def has_capability(self, capability: Capability | str) -> bool:
        """檢查設備是否具備指定能力"""
        name = capability.name if isinstance(capability, Capability) else capability
        return name in self._capability_bindings

    def get_binding(self, capability: Capability | str) -> CapabilityBinding | None:
        """取得能力綁定，不存在回傳 None"""
        name = capability.name if isinstance(capability, Capability) else capability
        return self._capability_bindings.get(name)

    def resolve_point(self, capability: Capability | str, slot: str) -> str:
        """
        解析能力的語意插槽到實際點位名稱

        Args:
            capability: 能力定義或名稱
            slot: 語意插槽名稱

        Returns:
            實際點位名稱

        Raises:
            ConfigurationError: 設備不具備該能力
            KeyError: slot 不存在
        """
        binding = self.get_binding(capability)
        if binding is None:
            name = capability.name if isinstance(capability, Capability) else capability
            raise ConfigurationError(f"Device '{self.device_id}' has no capability '{name}'")
        return binding.resolve(slot)

    def add_capability(self, binding: CapabilityBinding) -> None:
        """動態新增能力綁定（執行期）"""
        self._capability_bindings[binding.capability.name] = binding
        self._emitter.emit(
            EVENT_CAPABILITY_ADDED,
            CapabilityChangedPayload(device_id=self._config.device_id, capability_name=binding.capability.name),
        )

    def remove_capability(self, capability: Capability | str) -> None:
        """動態移除能力綁定（執行期）"""
        name = capability.name if isinstance(capability, Capability) else capability
        if name in self._capability_bindings:
            self._capability_bindings.pop(name, None)
            self._emitter.emit(
                EVENT_CAPABILITY_REMOVED,
                CapabilityChangedPayload(device_id=self._config.device_id, capability_name=name),
            )

    # =============== Point Toggle ===============

    def enable_point(self, name: str) -> None:
        """啟用點位"""
        if name not in self.all_point_names:
            raise KeyError(f"點位 '{name}' 不存在")
        self._disabled_points.discard(name)
        self._emitter.emit(
            EVENT_POINT_TOGGLED,
            PointToggledPayload(device_id=self._config.device_id, point_name=name, enabled=True),
        )

    def disable_point(self, name: str) -> None:
        """停用點位（讀取值不更新、告警不評估、寫入被拒）"""
        if name not in self.all_point_names:
            raise KeyError(f"點位 '{name}' 不存在")
        self._disabled_points.add(name)
        self._emitter.emit(
            EVENT_POINT_TOGGLED,
            PointToggledPayload(device_id=self._config.device_id, point_name=name, enabled=False),
        )

    def is_point_enabled(self, name: str) -> bool:
        """檢查點位是否啟用"""
        return name not in self._disabled_points

    # =============== Point Query ===============

    @property
    def read_points(self) -> tuple[ReadPoint, ...]:
        """所有固定讀取點位"""
        return self._read_points_always

    @property
    def rotating_read_points(self) -> tuple[tuple[ReadPoint, ...], ...]:
        """所有輪替讀取點位"""
        return self._read_points_rotating

    @property
    def write_point_names(self) -> list[str]:
        """所有寫入點位名稱"""
        return list(self._write_points.keys())

    @property
    def all_point_names(self) -> set[str]:
        """所有點位名稱（讀+寫）"""
        names = {p.name for p in self._read_points_always}
        for group in self._read_points_rotating:
            names.update(p.name for p in group)
        names.update(self._write_points.keys())
        return names

    @property
    def disabled_points(self) -> frozenset[str]:
        """目前被停用的點位"""
        return frozenset(self._disabled_points)

    def get_point_info(self) -> list[PointInfo]:
        """取得所有點位的詳細資訊（含啟用狀態）"""
        infos: list[PointInfo] = []
        write_names = set(self._write_points.keys())

        # 讀取點位
        for point in self._read_points_always:
            direction = "read_write" if point.name in write_names else "read"
            infos.append(
                PointInfo(
                    name=point.name,
                    address=point.address,
                    data_type=point.data_type,
                    direction=direction,
                    enabled=point.name not in self._disabled_points,
                    read_group=point.read_group,
                    metadata=point.metadata,
                )
            )

        seen_names = {p.name for p in self._read_points_always}
        for group in self._read_points_rotating:
            for point in group:
                if point.name in seen_names:
                    continue
                seen_names.add(point.name)
                direction = "read_write" if point.name in write_names else "read"
                infos.append(
                    PointInfo(
                        name=point.name,
                        address=point.address,
                        data_type=point.data_type,
                        direction=direction,
                        enabled=point.name not in self._disabled_points,
                        read_group=point.read_group,
                        metadata=point.metadata,
                    )
                )

        # 僅寫入的點位
        for name, wp in self._write_points.items():
            if name not in seen_names:
                infos.append(
                    PointInfo(
                        name=wp.name,
                        address=wp.address,
                        data_type=wp.data_type,
                        direction="write",
                        enabled=wp.name not in self._disabled_points,
                        read_group="",
                        metadata=wp.metadata,
                    )
                )

        return infos

    # =============== Reconfigure ===============

    async def reconfigure(self, spec: ReconfigureSpec) -> None:
        """
        動態重新配置點位

        Args:
            spec: 重新配置規格，None 欄位表示保持不變
        """
        was_running = self.is_running
        if was_running:
            await self.stop()

        changed_sections: list[str] = []

        try:
            if spec.always_points is not None or spec.rotating_points is not None:
                if spec.always_points is not None:
                    self._read_points_always = tuple(spec.always_points)
                    changed_sections.append("always_points")
                if spec.rotating_points is not None:
                    self._read_points_rotating = tuple(tuple(pts) for pts in spec.rotating_points)
                    changed_sections.append("rotating_points")
                self._scheduler.update_groups(
                    always_groups=self._grouper.group(list(self._read_points_always)),
                    rotating_groups=[self._grouper.group(list(pts)) for pts in self._read_points_rotating],
                )

            if spec.write_points is not None:
                self._write_points = {wp.name: wp for wp in spec.write_points}
                changed_sections.append("write_points")

            if spec.alarm_evaluators is not None:
                old_states = self._alarm_manager.export_states()
                self._alarm_manager = AlarmStateManager()
                self._alarm_evaluators = list(spec.alarm_evaluators)
                for evaluator in self._alarm_evaluators:
                    self._alarm_manager.register_alarms(evaluator.get_alarms())
                self._alarm_manager.import_states(old_states)
                changed_sections.append("alarm_evaluators")

            if spec.capability_bindings is not None:
                self._capability_bindings = {b.capability.name: b for b in spec.capability_bindings}
                changed_sections.append("capability_bindings")

            # 清理不再存在的 disabled_points
            valid_names = self.all_point_names
            self._disabled_points = self._disabled_points & valid_names
        finally:
            if was_running:
                await self.start()

        self._emitter.emit(
            EVENT_RECONFIGURED,
            ReconfiguredPayload(device_id=self._config.device_id, changed_sections=tuple(changed_sections)),
        )

    async def restart(self) -> None:
        """重啟讀取迴圈"""
        await self.stop()
        await self.start()
        self._emitter.emit(
            EVENT_RESTARTED,
            RestartedPayload(device_id=self._config.device_id),
        )

    # =============== Health ===============

    def health(self) -> HealthReport:
        """取得設備健康報告"""
        if self.is_healthy:
            status = HealthStatus.HEALTHY
        elif self.is_connected:
            status = HealthStatus.DEGRADED
        else:
            status = HealthStatus.UNHEALTHY
        return HealthReport(
            status=status,
            component=f"device:{self.device_id}",
            details={
                "connected": self.is_connected,
                "responsive": self.is_responsive,
                "protected": self.is_protected,
                "active_alarms": len(self.active_alarms),
            },
        )

    # =============== Lifecycle ===============

    async def connect(self) -> None:
        """
        連線設備

        Raises:
            DeviceConnectionError: 連線失敗
        """
        try:
            await self._client.connect()
        except Exception as e:
            raise DeviceConnectionError(self._config.device_id, f"連線失敗: {e}") from e
        await self._emitter.start()  # 啟動事件處理 worker
        async with self._status_lock:
            self._client_connected = True
            self._device_responsive = True  # 假設連線成功即可回應，讀取失敗會更新
            self._consecutive_failures = 0
        await self._emitter.emit_await(EVENT_CONNECTED, ConnectedPayload(device_id=self._config.device_id))

    async def disconnect(self) -> None:
        """
        斷線設備

        Raises:
            DeviceConnectionError: 斷線失敗
        """
        await self.stop()
        if self._client_connected:
            try:
                await self._client.disconnect()
            except Exception as e:
                raise DeviceConnectionError(self._config.device_id, f"斷線失敗: {e}") from e
            async with self._status_lock:
                self._client_connected = False
                self._device_responsive = False
                self._consecutive_failures = 0
                self._last_failure_time = None
            await self._emitter.emit_await(
                EVENT_DISCONNECTED,
                DisconnectPayload(device_id=self._config.device_id, reason="normal", consecutive_failures=0),
            )
            await self._emitter.stop()  # 停止事件處理 worker

    async def start(self) -> None:
        """
        啟動定期讀取循環

        即使未連線也可啟動，read_loop 會自動嘗試連線。
        """
        if self._read_task is not None and not self._read_task.done():
            return

        # 確保 emitter 已啟動
        await self._emitter.start()

        self._stop_event.clear()
        self._read_task = asyncio.create_task(self._read_loop())

    async def stop(self) -> None:
        """
        停止定期讀取循環
        """
        await self.cancel_pending_pulses()
        self._stop_event.set()
        if self._read_task is not None and not self._read_task.done():
            self._read_task.cancel()
            try:
                await self._read_task
            except asyncio.CancelledError:
                pass
            self._read_task = None

    async def __aenter__(self) -> AsyncModbusDevice:
        try:
            await self.connect()
            await self.start()
            return self
        except Exception:
            await self.disconnect()  # 清理
            raise

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            await self.stop()
        finally:
            await self.disconnect()

    # =============== Transport ===============

    async def read_once(self) -> dict[str, Any]:
        """
        執行一次完整的讀取流程

        包含：連線檢查（自動重連）、讀取點位、更新狀態、處理值變更事件、評估告警。
        適合不需要定期讀取的場景（如：手動觸發讀取、群組順序讀取）。

        Returns:
            讀取到的點位值字典

        Raises:
            Exception: 讀取失敗時拋出例外（會發送 EVENT_READ_ERROR）
        """
        # 未連線時嘗試重連
        if not self._client_connected:
            try:
                await self._client.connect()
                async with self._status_lock:
                    self._client_connected = True
                    # 注意：_device_responsive 保持 False，等讀取成功後才設為 True 並發送 EVENT_CONNECTED
                    self._consecutive_failures = 0
            except DeviceConnectionError:
                async with self._status_lock:
                    self._last_failure_time = time.monotonic()
                raise
            except Exception as e:
                async with self._status_lock:
                    self._last_failure_time = time.monotonic()
                raise DeviceConnectionError(self._config.device_id, f"重連失敗: {e}") from e

        start_time = time.monotonic()

        try:
            values = await self._read_all()

            # 更新狀態（成功路徑）
            async with self._status_lock:
                self._consecutive_failures = 0
                self._last_failure_time = None
                was_unresponsive = not self._device_responsive
                if was_unresponsive:
                    self._device_responsive = True

            # 設備恢復回應（在鎖外發送事件，避免死鎖）
            if was_unresponsive:
                await self._emitter.emit_await(
                    EVENT_CONNECTED,
                    ConnectedPayload(device_id=self._config.device_id),
                )

            # 更新值並發送變更事件
            await self._process_values(values)

            # 執行告警評估
            await self._evaluate_alarm(values)

            duration_ms = (time.monotonic() - start_time) * 1000
            self._emitter.emit(
                EVENT_READ_COMPLETE,
                ReadCompletePayload(
                    device_id=self._config.device_id,
                    values=values,
                    duration_ms=duration_ms,
                ),
            )

            return values

        except CommunicationError as e:
            # 已有正確 device_id 的 CommunicationError 直接傳播
            if e.device_id != "unknown":
                await self._handle_read_failure(str(e))
                await self._check_disconnect_threshold(str(e))
                raise
            # transport 層的 "unknown" device_id → 替換為正確 device_id
            err = CommunicationError(self._config.device_id, str(e))
            err.__cause__ = e.__cause__
            await self._handle_read_failure(str(err))
            await self._check_disconnect_threshold(str(err))
            raise err from e.__cause__
        except Exception as e:
            err = CommunicationError(self._config.device_id, f"讀取失敗: {e}")
            await self._handle_read_failure(str(err))
            await self._check_disconnect_threshold(str(err))
            raise err from e

    # =============== Events ================

    def on(self, event: str, handler: AsyncHandler) -> Callable[[], None]:
        """註冊事件處理器"""
        return self._emitter.on(event, handler)

    def emit(self, event: str, payload: Any) -> None:
        """發送事件（非阻塞）"""
        self._emitter.emit(event, payload)

    # =============== Private ===============

    async def _read_all(self) -> dict[str, Any]:
        """讀取所有點位（使用排程器預計算分組）"""
        groups = self._scheduler.get_next_groups()
        if not groups:
            return {}
        raw_values = await self._reader.read_many(groups)
        if self._aggregator_pipeline:
            return self._aggregator_pipeline.process(raw_values)
        return raw_values

    async def _read_loop(self) -> None:
        """讀取循環（含自動重連）"""
        interval = self._config.read_interval
        reconnect_interval = self._config.reconnect_interval

        while not self._stop_event.is_set():
            start_time = time.monotonic()

            # 未連線時嘗試重連
            if not self._client_connected:
                try:
                    await self._client.connect()
                    async with self._status_lock:
                        self._client_connected = True
                        self._device_responsive = True
                        self._consecutive_failures = 0
                    await self._emitter.emit_await(EVENT_CONNECTED, ConnectedPayload(device_id=self._config.device_id))
                except DeviceConnectionError:
                    # 重連失敗，等待後重試
                    await asyncio.sleep(reconnect_interval)
                    continue
                except Exception as e:
                    # 重連失敗，等待後重試
                    logger.warning(f"[{self._config.device_id}] Reconnect failed: {e}")
                    await asyncio.sleep(reconnect_interval)
                    continue

            try:
                await self.read_once()
            except Exception as e:
                logger.warning(f"[{self._config.device_id}] Read loop error: {e}")

            elapsed = time.monotonic() - start_time
            sleep_time = max(0, interval - elapsed)
            await asyncio.sleep(sleep_time)

    async def _handle_read_failure(self, error_msg: str) -> None:
        """處理讀取失敗：累加計數 + 記錄失敗時間 + 發送錯誤事件"""
        async with self._status_lock:
            self._consecutive_failures += 1
            self._last_failure_time = time.monotonic()
            failures = self._consecutive_failures
        self._emitter.emit(
            EVENT_READ_ERROR,
            ReadErrorPayload(
                device_id=self._config.device_id,
                error=error_msg,
                consecutive_failures=failures,
            ),
        )

    async def _check_disconnect_threshold(self, error_msg: str) -> None:
        """達到斷線閾值時標記設備無回應"""
        failures = 0
        async with self._status_lock:
            should_disconnect = (
                self._consecutive_failures >= self._config.disconnect_threshold and self._device_responsive
            )
            if should_disconnect:
                self._device_responsive = False
                failures = self._consecutive_failures

        if should_disconnect:
            await self._emitter.emit_await(
                EVENT_DISCONNECTED,
                DisconnectPayload(
                    device_id=self._config.device_id,
                    reason=error_msg,
                    consecutive_failures=failures,
                ),
            )

    async def _process_values(self, values: dict[str, Any]) -> None:
        """處理讀取到的值，發送變更事件（跳過 disabled 點位）"""
        for name, new_value in values.items():
            if name in self._disabled_points:
                continue
            try:
                old_value = self._latest_values.get(name)
                if old_value != new_value:
                    self._emitter.emit(
                        EVENT_VALUE_CHANGE,
                        ValueChangePayload(
                            device_id=self._config.device_id,
                            point_name=name,
                            old_value=old_value,
                            new_value=new_value,
                        ),
                    )
                self._latest_values[name] = new_value
            except Exception as e:
                logger.warning(f"[{self._config.device_id}] Event processing failed for point '{name}': {e}")

    # =============== Magic Methods =========

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} id={self.device_id} connected={self.is_connected} responsive={self.is_responsive}>"

    def __repr__(self) -> str:
        return self.__str__()



# =============== Load Shedding Strategy ===============
#
# 階段性負載卸載策略
#
# 在離網 (islanding) 等場景中，根據電池 SOC、剩餘時間等條件，
# 按優先順序逐步卸載/恢復負載，延長供電時間。
#
# Protocol:
#   - LoadCircuitProtocol: 負載迴路控制協定
#   - ShedCondition: 卸載觸發條件協定
#
# 內建條件:
#   - ThresholdCondition: 閾值條件
#   - RemainingTimeCondition: 剩餘時間條件
#
# Strategy:
#   - LoadSheddingStrategy: 階段性負載卸載策略

from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass, field
from typing import Protocol, runtime_checkable

from csp_lib.controller.core import (
    Command,
    ConfigMixin,
    ExecutionConfig,
    ExecutionMode,
    Strategy,
    StrategyContext,
)
from csp_lib.core import get_logger

logger = get_logger(__name__)


# =============== Protocols ===============


@runtime_checkable
class LoadCircuitProtocol(Protocol):
    """負載迴路控制協定"""

    @property
    def name(self) -> str: ...

    @property
    def is_shed(self) -> bool: ...

    async def shed(self) -> None: ...

    async def restore(self) -> None: ...


@runtime_checkable
class ShedCondition(Protocol):
    """卸載觸發條件"""

    def should_shed(self, context: StrategyContext) -> bool: ...

    def should_restore(self, context: StrategyContext) -> bool: ...


# =============== Built-in Conditions ===============


class ThresholdCondition:
    """
    閾值條件

    當 context.extra[context_key] < shed_below 時觸發卸載，
    當 context.extra[context_key] > restore_above 時觸發恢復。
    """

    def __init__(self, context_key: str, shed_below: float, restore_above: float) -> None:
        if restore_above < shed_below:
            raise ValueError(f"restore_above ({restore_above}) must be >= shed_below ({shed_below})")
        self._context_key = context_key
        self._shed_below = shed_below
        self._restore_above = restore_above

    def should_shed(self, context: StrategyContext) -> bool:
        value = context.extra.get(self._context_key)
        if value is None:
            return False
        return value < self._shed_below

    def should_restore(self, context: StrategyContext) -> bool:
        value = context.extra.get(self._context_key)
        if value is None:
            return False
        return value > self._restore_above


class RemainingTimeCondition:
    """
    剩餘時間條件

    當電池剩餘時間低於 shed_below 分鐘時觸發卸載，
    高於 restore_above 分鐘時觸發恢復。
    """

    def __init__(
        self,
        context_key: str = "battery_remaining_minutes",
        shed_below: float = 30.0,
        restore_above: float = 45.0,
    ) -> None:
        if restore_above < shed_below:
            raise ValueError(f"restore_above ({restore_above}) must be >= shed_below ({shed_below})")
        self._context_key = context_key
        self._shed_below = shed_below
        self._restore_above = restore_above

    def should_shed(self, context: StrategyContext) -> bool:
        value = context.extra.get(self._context_key)
        if value is None:
            return False
        return value < self._shed_below

    def should_restore(self, context: StrategyContext) -> bool:
        value = context.extra.get(self._context_key)
        if value is None:
            return False
        return value > self._restore_above


# =============== Shed Stage ===============


@dataclass(frozen=True)
class ShedStage:
    """
    卸載階段

    Attributes:
        name: 階段名稱
        circuits: 此階段要卸載的迴路
        condition: 觸發條件
        priority: 越小越先卸（越晚恢復）
        min_hold_seconds: 最小保持時間（防抖動）
    """

    name: str
    circuits: list[LoadCircuitProtocol]
    condition: ShedCondition
    priority: int = 0
    min_hold_seconds: float = 30.0


# =============== Config ===============


@dataclass
class LoadSheddingConfig(ConfigMixin):
    """
    負載卸載策略配置

    Attributes:
        stages: 卸載階段列表
        evaluation_interval: 評估週期（秒）
        restore_delay: 恢復延遲（秒，防止頻繁切換）
        auto_restore_on_deactivate: 策略停用時是否自動恢復所有負載
    """

    stages: list[ShedStage] = field(default_factory=list)
    evaluation_interval: int = 5
    restore_delay: float = 60.0
    auto_restore_on_deactivate: bool = True


# =============== Internal State ===============


class _StageState:
    """單一階段的內部狀態"""

    __slots__ = ("is_shed", "shed_at", "restore_requested_at")

    def __init__(self) -> None:
        self.is_shed: bool = False
        self.shed_at: float = 0.0
        self.restore_requested_at: float | None = None


# =============== Strategy ===============


class LoadSheddingStrategy(Strategy):
    """
    階段性負載卸載策略

    - PERIODIC 模式持續評估條件
    - execute() 純條件判斷，記錄需要 shed/restore 的 stages
    - 背景 Task 執行實際的 async 斷路器操作
    - 卸載順序：priority 升序（低優先級先卸）
    - 恢復順序：priority 降序（高優先級先恢復）
    """

    def __init__(self, config: LoadSheddingConfig) -> None:
        self._config = config
        self._sorted_stages = sorted(config.stages, key=lambda s: s.priority)
        self._states: dict[str, _StageState] = {stage.name: _StageState() for stage in config.stages}
        self._pending_actions: list[tuple[str, str]] = []  # (stage_name, "shed"|"restore")
        self._action_task: asyncio.Task[None] | None = None
        self._action_event = asyncio.Event()

    @property
    def config(self) -> LoadSheddingConfig:
        return self._config

    @property
    def execution_config(self) -> ExecutionConfig:
        return ExecutionConfig(mode=ExecutionMode.PERIODIC, interval_seconds=self._config.evaluation_interval)

    def execute(self, context: StrategyContext) -> Command:
        now = time.monotonic()

        # 卸載評估（priority 升序：低優先級先卸）
        for stage in self._sorted_stages:
            state = self._states[stage.name]
            if not state.is_shed and stage.condition.should_shed(context):
                state.is_shed = True
                state.shed_at = now
                state.restore_requested_at = None
                self._pending_actions.append((stage.name, "shed"))
                logger.info(f"Load shedding: stage '{stage.name}' scheduled for shed")

        # 恢復評估（priority 降序：高優先級先恢復）
        for stage in reversed(self._sorted_stages):
            state = self._states[stage.name]
            if not state.is_shed:
                continue
            if not stage.condition.should_restore(context):
                state.restore_requested_at = None
                continue

            # 檢查 min_hold_seconds
            if (now - state.shed_at) < stage.min_hold_seconds:
                continue

            # 開始 restore delay
            if state.restore_requested_at is None:
                state.restore_requested_at = now
                logger.debug(f"Load shedding: stage '{stage.name}' restore delay started")
                continue

            # 檢查 restore_delay
            if (now - state.restore_requested_at) >= self._config.restore_delay:
                state.is_shed = False
                state.restore_requested_at = None
                self._pending_actions.append((stage.name, "restore"))
                logger.info(f"Load shedding: stage '{stage.name}' scheduled for restore")

        if self._pending_actions:
            self._action_event.set()

        return context.last_command

    async def on_activate(self) -> None:
        self._action_event.clear()
        self._action_task = asyncio.create_task(self._action_loop())
        logger.info("LoadSheddingStrategy activated")

    async def on_deactivate(self) -> None:
        if self._action_task is not None:
            self._action_task.cancel()
            try:
                await self._action_task
            except asyncio.CancelledError:
                pass
            self._action_task = None

        if self._config.auto_restore_on_deactivate:
            await self._restore_all()

        # 重置狀態
        self._pending_actions.clear()
        for state in self._states.values():
            state.is_shed = False
            state.shed_at = 0.0
            state.restore_requested_at = None
        logger.info("LoadSheddingStrategy deactivated")

    @property
    def shed_stage_names(self) -> list[str]:
        """當前已卸載的階段名稱"""
        return [name for name, state in self._states.items() if state.is_shed]

    async def _action_loop(self) -> None:
        """背景 task：執行待處理的 shed/restore 操作"""
        while True:
            await self._action_event.wait()
            self._action_event.clear()

            actions = list(self._pending_actions)
            self._pending_actions.clear()

            stage_map = {s.name: s for s in self._config.stages}
            for stage_name, action in actions:
                stage = stage_map.get(stage_name)
                if stage is None:
                    continue
                for circuit in stage.circuits:
                    try:
                        if action == "shed":
                            await circuit.shed()
                            logger.info(f"Circuit '{circuit.name}' shed (stage: {stage_name})")
                        else:
                            await circuit.restore()
                            logger.info(f"Circuit '{circuit.name}' restored (stage: {stage_name})")
                    except Exception:
                        logger.opt(exception=True).warning(
                            f"Failed to {action} circuit '{circuit.name}' (stage: {stage_name})"
                        )

    async def _restore_all(self) -> None:
        """恢復所有已卸載的迴路（priority 降序）"""
        for stage in reversed(self._sorted_stages):
            state = self._states[stage.name]
            if not state.is_shed:
                continue
            for circuit in stage.circuits:
                try:
                    if circuit.is_shed:
                        await circuit.restore()
                        logger.info(f"Auto-restore circuit '{circuit.name}' (stage: {stage.name})")
                except Exception:
                    logger.opt(exception=True).warning(
                        f"Failed to auto-restore circuit '{circuit.name}' (stage: {stage.name})"
                    )

    def __str__(self) -> str:
        shed_count = sum(1 for s in self._states.values() if s.is_shed)
        return f"LoadSheddingStrategy(stages={len(self._config.stages)}, shed={shed_count})"



"""FastAPI dependency injection helpers."""

from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import Request

if TYPE_CHECKING:
    from csp_lib.controller.system import ModeManager
    from csp_lib.integration import DeviceRegistry, SystemController

    from .ws.manager import WebSocketManager


def get_system_controller(request: Request) -> SystemController:
    """Extract SystemController from app state."""
    return request.app.state.system_controller


def get_registry(request: Request) -> DeviceRegistry:
    """Extract DeviceRegistry from app state."""
    return request.app.state.system_controller.registry


def get_mode_manager(request: Request) -> ModeManager:
    """Extract ModeManager from app state."""
    return request.app.state.system_controller.mode_manager


def get_ws_manager(request: Request) -> WebSocketManager:
    """Extract WebSocketManager from app state."""
    return request.app.state.ws_manager



# =============== Modbus Exceptions ===============
#
# Modbus 模組例外類型定義
#
# 例外繼承結構：
#   ModbusError (基礎)
#   ├── ModbusEncodeError (編碼錯誤)
#   ├── ModbusDecodeError (解碼錯誤)
#   ├── ModbusConfigError (設定錯誤)
#   ├── ModbusCircuitBreakerError (斷路器開啟)
#   └── ModbusQueueFullError (請求佇列已滿)

from __future__ import annotations


class ModbusError(Exception):
    """
    Modbus 基礎例外

    所有 Modbus 相關例外的父類別，方便統一捕捉。

    Attributes:
        address: Modbus 暫存器起始位址
        unit_id: 設備位址 (slave ID)
        function_code: Modbus 功能碼 (e.g., "FC03")
    """

    def __init__(
        self,
        message: str = "",
        *,
        address: int | None = None,
        unit_id: int | None = None,
        function_code: str | None = None,
    ) -> None:
        self.address = address
        self.unit_id = unit_id
        self.function_code = function_code
        context_parts: list[str] = []
        if address is not None:
            context_parts.append(f"addr={address}")
        if unit_id is not None:
            context_parts.append(f"unit={unit_id}")
        if function_code is not None:
            context_parts.append(f"fc={function_code}")
        full_msg = f"{message} [{', '.join(context_parts)}]" if context_parts else message
        super().__init__(full_msg)


class ModbusEncodeError(ModbusError):
    """
    編碼錯誤

    當編碼過程失敗時拋出，常見原因：
    - 值超出資料類型允許範圍 (e.g., int16 超過 32767)
    - 類型不符 (e.g., 傳入字串給數值類型)
    """


class ModbusDecodeError(ModbusError):
    """
    解碼錯誤

    當解碼過程失敗時拋出，常見原因：
    - 暫存器資料長度不足
    - 資料格式錯誤 (e.g., 無效的 IEEE 754 格式)
    """


class ModbusConfigError(ModbusError):
    """
    設定錯誤

    當設定參數無效時拋出，常見原因：
    - 無效的連線參數 (e.g., 負的 port 號)
    - 無效的資料類型參數 (e.g., bit_width 非 16 的倍數)
    """


class ModbusCircuitBreakerError(ModbusError):
    """
    斷路器開啟錯誤

    當某個 unit_id 的斷路器處於 OPEN 狀態時拋出，
    表示該設備連續失敗次數已達閾值，暫時停止請求。

    Attributes:
        unit_id: 觸發斷路器的設備位址
    """

    def __init__(self, unit_id: int, message: str | None = None) -> None:
        super().__init__(message or f"Circuit breaker is open for unit_id={unit_id}", unit_id=unit_id)


class ModbusQueueFullError(ModbusError):
    """
    請求佇列已滿錯誤

    當請求佇列達到 max_queue_size 上限時拋出。
    """


__all__ = [
    "ModbusError",
    "ModbusEncodeError",
    "ModbusDecodeError",
    "ModbusConfigError",
    "ModbusCircuitBreakerError",
    "ModbusQueueFullError",
]



# =============== Context Module ===============
#
# 策略執行時上下文
# - StrategyContext: 執行時上下文

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional

from .command import Command, SystemBase

if TYPE_CHECKING:
    from csp_lib.core.runtime_params import RuntimeParameters


@dataclass
class StrategyContext:
    """
    策略執行時上下文 (唯讀)

    提供策略執行所需的外部狀態，由 Executor 注入。
    策略不應直接修改此物件。

    Attributes:
        last_command: 上一次執行的命令
        soc: 儲能系統 SOC (%)
        current_time: 當前時間
        extra: 額外資料 — 設備讀值（frequency, meter_power 等，由 ContextMapping 注入）
        params: 系統參數 — RuntimeParameters 引用（EMS 指令、保護設定等，由 SystemController 注入）
    """

    last_command: Command = field(default_factory=Command)
    soc: Optional[float] = None
    system_base: Optional[SystemBase] = None
    current_time: Optional[datetime] = None
    extra: dict[str, Any] = field(default_factory=dict)
    params: RuntimeParameters | None = None

    def percent_to_kw(self, p_percent: float) -> float:
        """將百分比轉換為 kW"""
        if self.system_base is None:
            raise ValueError("system_base is not set in StrategyContext")
        return p_percent * self.system_base.p_base / 100

    def percent_to_kvar(self, q_percent: float) -> float:
        """將百分比轉換為 kVar"""
        if self.system_base is None:
            raise ValueError("system_base is not set in StrategyContext")
        return q_percent * self.system_base.q_base / 100



# =============== Equipment IO - Reader ===============
#
# 群組讀取器
#
# 負責執行 Modbus 讀取並解碼群組資料

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Sequence

from csp_lib.core.errors import CommunicationError, ConfigurationError
from csp_lib.modbus.enums import FunctionCode
from csp_lib.modbus.exceptions import ModbusError

from .base import ReadGroup

if TYPE_CHECKING:
    from csp_lib.modbus import AsyncModbusClientBase


class GroupReader:
    """
    群組讀取器

    負責執行 Modbus 讀取並解碼群組資料。
    與 ValidatedWriter 對稱，提供完整的 I/O 讀寫能力。

    Attributes:
        client: Modbus 客戶端
        address_offset: 位址偏移（PLC 1-based: offset=1）
        max_concurrent_reads: 最大並行讀取數
            - TCP client: 預設 3（可同時多個請求）
            - SharedTCP/RTU client: 預設 1（串列讀取）

    使用範例：
        from csp_lib.equipment.transport import GroupReader, PointGrouper, ReadScheduler

        grouper = PointGrouper()
        scheduler = ReadScheduler(always_groups=grouper.group(points))

        # TCP 設備可並行讀取
        reader = GroupReader(client, max_concurrent_reads=3)

        # RTU/SharedTCP 設備需串列讀取
        reader = GroupReader(client, max_concurrent_reads=1)

        # 讀取下一批群組
        groups = scheduler.get_next_groups()
        data = await reader.read_many(groups)
    """

    def __init__(
        self,
        client: AsyncModbusClientBase,
        unit_id: int = 1,
        address_offset: int = 0,
        max_concurrent_reads: int = 1,
    ):
        """
        初始化群組讀取器

        Args:
            client: Modbus 客戶端
            unit_id: 設備位址 (Slave ID)
            address_offset: 位址偏移（PLC 1-based 定址時設為 1）
            max_concurrent_reads: 最大並行讀取數（預設 1 = 串列讀取）
        """
        if max_concurrent_reads < 1:
            raise ConfigurationError(f"max_concurrent_reads 必須 >= 1，收到: {max_concurrent_reads}")

        self._client = client
        self._unit_id = unit_id
        self._address_offset = address_offset
        self._max_concurrent_reads = max_concurrent_reads
        self._semaphore = asyncio.Semaphore(max_concurrent_reads)

    async def read(self, group: ReadGroup) -> dict[str, Any]:
        """
        讀取單一群組並解碼

        Args:
            group: 讀取群組

        Returns:
            {點位名稱: 值} 字典

        Raises:
            Exception: Modbus 通訊錯誤時傳播原始異常
        """
        async with self._semaphore:
            raw_data = await self._read_from_device(group)
            return self._decode(group, raw_data)

    async def read_many(self, groups: Sequence[ReadGroup]) -> dict[str, Any]:
        """
        讀取多個群組並合併結果

        若 max_concurrent_reads > 1，會並行讀取多個群組。

        Args:
            groups: 讀取群組列表

        Returns:
            合併的 {點位名稱: 值} 字典
        """
        if self._max_concurrent_reads == 1:
            # 串列讀取（RTU/SharedTCP）
            result: dict[str, Any] = {}
            for group in groups:
                data = await self.read(group)
                result.update(data)
            return result

        # 並行讀取（TCP）
        tasks = [self.read(group) for group in groups]
        results = await asyncio.gather(*tasks)

        merged: dict[str, Any] = {}
        for data in results:
            merged.update(data)
        return merged

    async def _read_from_device(self, group: ReadGroup) -> list[int] | list[bool]:
        """
        根據 function_code 讀取資料

        Args:
            group: 讀取群組

        Returns:
            原始暫存器/線圈資料

        Raises:
            ConfigurationError: 不支援的 FunctionCode
            CommunicationError: Modbus 通訊錯誤
        """
        address = group.start_address + self._address_offset
        count = group.count
        function_code = group.function_code

        try:
            if function_code == FunctionCode.READ_COILS:
                return list(await self._client.read_coils(address, count, self._unit_id))
            elif function_code == FunctionCode.READ_DISCRETE_INPUTS:
                return list(await self._client.read_discrete_inputs(address, count, self._unit_id))
            elif function_code == FunctionCode.READ_HOLDING_REGISTERS:
                return list(await self._client.read_holding_registers(address, count, self._unit_id))
            elif function_code == FunctionCode.READ_INPUT_REGISTERS:
                return list(await self._client.read_input_registers(address, count, self._unit_id))
            else:
                raise ConfigurationError(f"不支援的 Function Code: {function_code}")
        except (ConfigurationError, CommunicationError):
            raise
        except ModbusError as e:
            raise CommunicationError("unknown", f"Modbus 通訊錯誤: {e}") from e

    def _decode(self, group: ReadGroup, raw_data: list[int] | list[bool]) -> dict[str, Any]:
        """
        解碼群組讀取結果

        Args:
            group: 讀取群組
            raw_data: 原始資料

        Returns:
            {點位名稱: 值} 字典

        Raises:
            CommunicationError: 資料長度不足以解碼點位
        """
        result: dict[str, Any] = {}

        for point in group.points:
            # 計算偏移
            offset = point.address - group.start_address
            length = point.data_type.register_count

            # 提取資料切片並驗證長度
            data_slice = list(raw_data[offset : offset + length])
            if len(data_slice) < length:
                raise CommunicationError(
                    "unknown",
                    f"資料不足以解碼點位 '{point.name}': "
                    f"期望 {length} 個暫存器，實際 {len(data_slice)} "
                    f"(offset={offset}, group.count={group.count})",
                )

            # 解碼
            value = point.data_type.decode(
                registers=data_slice,
                byte_order=point.byte_order,
                register_order=point.register_order,
            )

            if point.pipeline:
                value = point.pipeline.process(value)

            result[point.name] = value

        return result


__all__ = [
    "GroupReader",
]



"""
MongoBatchUploader 設定模組

定義上傳器的配置參數，使用 frozen dataclass 確保不可變性。
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class UploaderConfig:
    """
    批次上傳器設定

    Attributes:
        flush_interval: 定期 flush 間隔（秒），預設 5 秒
        batch_size_threshold: 單一 collection 累積幾筆後觸發上傳，預設 100 筆
        max_queue_size: Queue 上限，超過後丟棄最舊資料，預設 10000 筆
        max_retry_count: 單批資料最大重試次數，超過後丟棄，預設 3 次
    """

    flush_interval: int = 5
    batch_size_threshold: int = 100
    max_queue_size: int = 10000
    max_retry_count: int = 3

    def __post_init__(self) -> None:
        """驗證設定值的合理性"""
        if self.flush_interval <= 0:
            raise ValueError(f"flush_interval 必須為正整數，收到: {self.flush_interval}")
        if self.batch_size_threshold <= 0:
            raise ValueError(f"batch_size_threshold 必須為正整數，收到: {self.batch_size_threshold}")
        if self.max_queue_size <= 0:
            raise ValueError(f"max_queue_size 必須為正整數，收到: {self.max_queue_size}")
        if self.max_retry_count < 0:
            raise ValueError(f"max_retry_count 不可為負數，收到: {self.max_retry_count}")



# =============== Hierarchical Status Report ===============
#
# 子執行器狀態回報資料結構
#
# 定義子站點向上層回報的狀態資訊：
#   - ExecutorStatus: 子執行器即時狀態
#   - StatusReport: 帶時間戳與站點標識的狀態封包

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any

from csp_lib.controller.core import Command


@dataclass(frozen=True)
class ExecutorStatus:
    """
    子執行器即時狀態

    描述單一站點控制器的當前運作狀態，供上層編排器決策。

    Attributes:
        strategy_name: 當前生效策略名稱（如 "pq", "cascading(pq+qv)"）
        last_command: 最近一次執行的命令
        active_overrides: 當前啟用的覆蓋模式列表
        base_modes: 當前設定的基礎模式列表
        is_running: 控制器是否正在運行
        device_count: 管理的設備數量
        healthy_device_count: 健康設備數量
    """

    strategy_name: str = ""
    last_command: Command = field(default_factory=Command)
    active_overrides: tuple[str, ...] = ()
    base_modes: tuple[str, ...] = ()
    is_running: bool = False
    device_count: int = 0
    healthy_device_count: int = 0


@dataclass(frozen=True)
class StatusReport:
    """
    狀態回報封包

    子站點向上層控制器回報的完整狀態資訊，
    包含站點標識、時間戳、執行器狀態與額外指標。

    Attributes:
        site_id: 回報站點 ID
        status: 執行器即時狀態
        timestamp: 回報時間 (UTC)
        metrics: 額外指標（如 SOC、電壓、頻率等）
    """

    site_id: str
    status: ExecutorStatus
    timestamp: datetime = field(default_factory=lambda: datetime.now(timezone.utc))
    metrics: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        """序列化為字典"""
        return {
            "site_id": self.site_id,
            "status": {
                "strategy_name": self.status.strategy_name,
                "last_command": {
                    "p_target": self.status.last_command.p_target,
                    "q_target": self.status.last_command.q_target,
                },
                "active_overrides": list(self.status.active_overrides),
                "base_modes": list(self.status.base_modes),
                "is_running": self.status.is_running,
                "device_count": self.status.device_count,
                "healthy_device_count": self.status.healthy_device_count,
            },
            "timestamp": self.timestamp.isoformat(),
            "metrics": self.metrics,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> StatusReport:
        """從字典反序列化"""
        status_data = data.get("status", {})
        cmd_data = status_data.get("last_command", {})
        return cls(
            site_id=data["site_id"],
            status=ExecutorStatus(
                strategy_name=status_data.get("strategy_name", ""),
                last_command=Command(
                    p_target=float(cmd_data.get("p_target", 0.0)),
                    q_target=float(cmd_data.get("q_target", 0.0)),
                ),
                active_overrides=tuple(status_data.get("active_overrides", [])),
                base_modes=tuple(status_data.get("base_modes", [])),
                is_running=status_data.get("is_running", False),
                device_count=status_data.get("device_count", 0),
                healthy_device_count=status_data.get("healthy_device_count", 0),
            ),
            timestamp=(
                datetime.fromisoformat(data["timestamp"]) if "timestamp" in data else datetime.now(timezone.utc)
            ),
            metrics=data.get("metrics", {}),
        )


__all__ = [
    "ExecutorStatus",
    "StatusReport",
]



# =============== Manager - Data ===============
#
# 資料上傳管理模組
#
# 提供設備資料自動上傳功能：
#   - DataUploadManager: 訂閱設備事件並自動上傳資料至 MongoDB
#
# 使用方式：
#   1. 建立 MongoBatchUploader 實例
#   2. 建立 DataUploadManager 並注入 Uploader
#   3. 呼叫 subscribe() 訂閱 AsyncModbusDevice
#   4. 讀取資料自動上傳至指定 collection

from .upload import DataUploadManager

__all__ = [
    "DataUploadManager",
]



# =============== Integration - System Controller ===============
#
# 頂層系統控制器
#
# 結合 ModeManager、ProtectionGuard 與既有整合元件，提供：
#   - 模式管理：註冊多種模式並以優先權切換
#   - 保護機制：SOC 限制、逆送保護、系統告警保護
#   - 自動告警處理：設備告警時自動推入 stop override
#
# 架構：
#   ContextBuilder.build() → StrategyContext (注入 system_alarm)
#        ↓
#   StrategyExecutor (策略由 ModeManager 決定)
#        ↓
#   Command (原始) → ProtectionGuard.apply() → Command (保護後)
#        ↓
#   CommandRouter.route() → 設備寫入

from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Awaitable, Callable

from csp_lib.controller.core import (
    Command,
    CommandProcessor,
    ExecutionConfig,
    ExecutionMode,
    StrategyContext,
    SystemBase,
)
from csp_lib.controller.executor import StrategyExecutor
from csp_lib.controller.services import PVDataService
from csp_lib.controller.strategies import StopStrategy
from csp_lib.controller.system import ModeManager, ModePriority, ProtectionGuard, ProtectionResult, ProtectionRule
from csp_lib.controller.system.cascading import CapacityConfig, CascadingStrategy
from csp_lib.controller.system.event_override import AlarmStopOverride, EventDrivenOverride
from csp_lib.controller.system.mode import SwitchSource
from csp_lib.core import AsyncLifecycleMixin, get_logger
from csp_lib.core.errors import ConfigurationError
from csp_lib.core.health import HealthReport, HealthStatus

from .command_router import CommandRouter
from .context_builder import ContextBuilder
from .data_feed import DeviceDataFeed
from .distributor import DeviceSnapshot, PowerDistributor
from .heartbeat import HeartbeatService
from .orchestrator import SystemCommandOrchestrator
from .registry import DeviceRegistry
from .schema import (
    CapabilityCommandMapping,
    CapabilityContextMapping,
    CapabilityRequirement,
    CommandMapping,
    ContextMapping,
    DataFeedMapping,
    HeartbeatMapping,
    HeartbeatMode,
)

if TYPE_CHECKING:
    from csp_lib.controller.core import Strategy
    from csp_lib.equipment.device import AsyncModbusDevice

logger = get_logger(__name__)

_AUTO_STOP_MODE = "__auto_stop__"
_SCHEDULE_MODE = "__schedule__"


@dataclass
class SystemControllerConfig:
    """
    SystemController 配置

    Attributes:
        context_mappings: 設備點位 → StrategyContext 的映射列表
        command_mappings: Command 欄位 → 設備寫入的映射列表
        capability_context_mappings: capability-driven context 映射列表
        capability_command_mappings: capability-driven command 映射列表
        system_base: 系統基準值（可選）
        data_feed_mapping: PV 資料餵入映射（可選）
        pv_max_history: PVDataService 最大歷史記錄數
        protection_rules: 保護規則列表
        auto_stop_on_alarm: 系統告警時自動推入 stop override
        system_alarm_key: system_alarm 在 context.extra 中的 key
        capacity_kva: 級聯策略最大視在功率 (kVA)（可選，多 base mode 時啟用）
        alarm_mode: 告警模式 - "system_wide" 全系統停機 / "per_device" 僅告警設備關機
        on_device_alarm: 設備告警時的回呼（per_device 模式）
        on_device_alarm_clear: 設備告警解除時的回呼（per_device 模式）
        heartbeat_mappings: 心跳寫入映射列表（設定時自動建立 HeartbeatService）
        heartbeat_interval: 心跳寫入間隔（秒），預設 1.0
        use_heartbeat_capability: 是否啟用 HEARTBEAT 能力發現模式
        heartbeat_capability_mode: 能力發現模式的心跳值模式
        heartbeat_capability_constant_value: CONSTANT 模式的固定寫入值
        heartbeat_capability_increment_max: INCREMENT 模式的最大計數值
        power_distributor: 功率分配器（可選），設定後 capability command mappings 使用 per-device 分配
        capability_requirements: 能力需求列表，供 preflight_check 驗證
        strict_capability_check: 啟用嚴格能力檢查（preflight 失敗時 raise ConfigurationError）
    """

    context_mappings: list[ContextMapping] = field(default_factory=list)
    command_mappings: list[CommandMapping] = field(default_factory=list)
    capability_context_mappings: list[CapabilityContextMapping] = field(default_factory=list)
    capability_command_mappings: list[CapabilityCommandMapping] = field(default_factory=list)
    system_base: SystemBase | None = None
    data_feed_mapping: DataFeedMapping | None = None
    pv_max_history: int = 300
    protection_rules: list[ProtectionRule] = field(default_factory=list)
    auto_stop_on_alarm: bool = True
    system_alarm_key: str = "system_alarm"
    capacity_kva: float | None = None
    alarm_mode: str = "system_wide"
    on_device_alarm: Callable[[AsyncModbusDevice], Awaitable[None]] | None = None
    on_device_alarm_clear: Callable[[AsyncModbusDevice], Awaitable[None]] | None = None
    heartbeat_mappings: list[HeartbeatMapping] = field(default_factory=list)
    heartbeat_interval: float = 1.0
    use_heartbeat_capability: bool = False
    heartbeat_capability_mode: HeartbeatMode = HeartbeatMode.TOGGLE
    heartbeat_capability_constant_value: int = 1
    heartbeat_capability_increment_max: int = 65535
    power_distributor: PowerDistributor | None = None
    post_protection_processors: list[CommandProcessor] = field(default_factory=list)
    runtime_params: Any = None  # RuntimeParameters | None, 自動注入到 StrategyContext.params
    capability_requirements: list[CapabilityRequirement] = field(default_factory=list)
    strict_capability_check: bool = False

    @classmethod
    def builder(cls) -> "SystemControllerConfigBuilder":
        """回傳 fluent builder 以逐步建構配置"""
        return SystemControllerConfigBuilder()


class SystemControllerConfigBuilder:
    """
    SystemControllerConfig 的 Fluent Builder

    透過鏈式呼叫逐步建構配置，最後 build() 回傳 frozen 的 config。

    Usage::

        config = (
            SystemControllerConfig.builder()
            .system_base(p_base=2000)
            .map_context(device_id="MTD1", point_name="f", target="extra.frequency")
            .map_context(trait="bms", point_name="soc", target="soc")
            .map_command(field="p_target", device_id="PCS1", point_name="set_p")
            .protect(DynamicSOCProtection(params))
            .processor(compensator)
            .params(runtime_params)
            .build()
        )
    """

    def __init__(self) -> None:
        self._context_mappings: list[ContextMapping] = []
        self._command_mappings: list[CommandMapping] = []
        self._cap_context_mappings: list[CapabilityContextMapping] = []
        self._cap_command_mappings: list[CapabilityCommandMapping] = []
        self._system_base: SystemBase | None = None
        self._data_feed_mapping: DataFeedMapping | None = None
        self._pv_max_history: int = 300
        self._protection_rules: list[ProtectionRule] = []
        self._auto_stop_on_alarm: bool = True
        self._system_alarm_key: str = "system_alarm"
        self._capacity_kva: float | None = None
        self._alarm_mode: str = "system_wide"
        self._on_device_alarm: Callable | None = None
        self._on_device_alarm_clear: Callable | None = None
        self._heartbeat_mappings: list[HeartbeatMapping] = []
        self._heartbeat_interval: float = 1.0
        self._use_heartbeat_capability: bool = False
        self._heartbeat_capability_mode: HeartbeatMode = HeartbeatMode.TOGGLE
        self._heartbeat_capability_constant_value: int = 1
        self._heartbeat_capability_increment_max: int = 65535
        self._power_distributor: PowerDistributor | None = None
        self._processors: list[CommandProcessor] = []
        self._runtime_params: Any = None
        self._capability_requirements: list[CapabilityRequirement] = []
        self._strict_capability_check: bool = False

    # ─────────────── 系統基準 ───────────────

    def system_base(self, p_base: float, q_base: float = 0.0) -> "SystemControllerConfigBuilder":
        """設定系統基準值 (kW / kVar)"""
        self._system_base = SystemBase(p_base=p_base, q_base=q_base)
        return self

    # ─────────────── Context Mapping ───────────────

    def map_context(
        self,
        point_name: str,
        target: str,
        *,
        device_id: str | None = None,
        trait: str | None = None,
        aggregate: Any = None,
        default: Any = None,
        transform: Callable | None = None,
    ) -> "SystemControllerConfigBuilder":
        """
        新增 context mapping（設備讀值 → StrategyContext）

        Args:
            point_name: 設備點位名稱
            target: context 欄位（如 "soc", "extra.frequency"）
            device_id: 指定設備（與 trait 二擇一）
            trait: 設備群組（與 device_id 二擇一）
            aggregate: 多設備聚合函式（trait 模式用）
            default: 無值時的預設
            transform: 值轉換函式
        """
        kwargs: dict[str, Any] = {"point_name": point_name, "context_field": target}
        if device_id is not None:
            kwargs["device_id"] = device_id
        if trait is not None:
            kwargs["trait"] = trait
        if aggregate is not None:
            kwargs["aggregate"] = aggregate
        if default is not None:
            kwargs["default"] = default
        if transform is not None:
            kwargs["transform"] = transform
        self._context_mappings.append(ContextMapping(**kwargs))
        return self

    # ─────────────── Command Mapping ───────────────

    def map_command(
        self,
        field: str,
        point_name: str,
        *,
        device_id: str | None = None,
        trait: str | None = None,
        transform: Callable | None = None,
    ) -> "SystemControllerConfigBuilder":
        """
        新增 command mapping（Command → 設備寫入）

        Args:
            field: Command 欄位（如 "p_target", "q_target"）
            point_name: 設備寫入點位名稱
            device_id: 指定設備（與 trait 二擇一）
            trait: 設備群組（與 device_id 二擇一）
            transform: 值轉換函式
        """
        kwargs: dict[str, Any] = {"command_field": field, "point_name": point_name}
        if device_id is not None:
            kwargs["device_id"] = device_id
        if trait is not None:
            kwargs["trait"] = trait
        if transform is not None:
            kwargs["transform"] = transform
        self._command_mappings.append(CommandMapping(**kwargs))
        return self

    # ─────────────── Capability Mapping ───────────────

    def map_capability_context(self, mapping: CapabilityContextMapping) -> "SystemControllerConfigBuilder":
        """新增 capability context mapping"""
        self._cap_context_mappings.append(mapping)
        return self

    def map_capability_command(self, mapping: CapabilityCommandMapping) -> "SystemControllerConfigBuilder":
        """新增 capability command mapping"""
        self._cap_command_mappings.append(mapping)
        return self

    # ─────────────── 保護 ───────────────

    def protect(self, rule: ProtectionRule) -> "SystemControllerConfigBuilder":
        """新增保護規則"""
        self._protection_rules.append(rule)
        return self

    def auto_stop(self, enabled: bool = True, alarm_key: str = "system_alarm") -> "SystemControllerConfigBuilder":
        """設定自動停機"""
        self._auto_stop_on_alarm = enabled
        self._system_alarm_key = alarm_key
        return self

    # ─────────────── Post-Protection Processor ───────────────

    def processor(self, proc: CommandProcessor) -> "SystemControllerConfigBuilder":
        """新增 post-protection 命令處理器（如 PowerCompensator）"""
        self._processors.append(proc)
        return self

    # ─────────────── RuntimeParameters ───────────────

    def params(self, runtime_params: Any) -> "SystemControllerConfigBuilder":
        """設定 RuntimeParameters（注入到 StrategyContext.params）"""
        self._runtime_params = runtime_params
        return self

    # ─────────────── 功率分配 ───────────────

    def distributor(self, dist: PowerDistributor) -> "SystemControllerConfigBuilder":
        """設定功率分配器"""
        self._power_distributor = dist
        return self

    # ─────────────── 心跳 ───────────────

    def heartbeat(
        self,
        mappings: list[HeartbeatMapping] | None = None,
        interval: float = 1.0,
        use_capability: bool = False,
        mode: HeartbeatMode = HeartbeatMode.TOGGLE,
    ) -> "SystemControllerConfigBuilder":
        """設定心跳服務"""
        if mappings:
            self._heartbeat_mappings = mappings
        self._heartbeat_interval = interval
        self._use_heartbeat_capability = use_capability
        self._heartbeat_capability_mode = mode
        return self

    # ─────────────── 告警模式 ───────────────

    def alarm_mode_per_device(
        self,
        on_alarm: Callable | None = None,
        on_clear: Callable | None = None,
    ) -> "SystemControllerConfigBuilder":
        """設定 per-device 告警模式"""
        self._alarm_mode = "per_device"
        self._on_device_alarm = on_alarm
        self._on_device_alarm_clear = on_clear
        return self

    # ─────────────── PV / 級聯 ───────────────

    def data_feed(self, mapping: DataFeedMapping, max_history: int = 300) -> "SystemControllerConfigBuilder":
        """設定 PV 資料餵入"""
        self._data_feed_mapping = mapping
        self._pv_max_history = max_history
        return self

    def cascading(self, capacity_kva: float) -> "SystemControllerConfigBuilder":
        """設定多 base mode 級聯的最大視在功率"""
        self._capacity_kva = capacity_kva
        return self

    # ─────────────── Capability Requirements ───────────────

    def require_capability(self, requirement: CapabilityRequirement) -> "SystemControllerConfigBuilder":
        """新增能力需求"""
        self._capability_requirements.append(requirement)
        return self

    def strict_capability(self, enabled: bool = True) -> "SystemControllerConfigBuilder":
        """啟用嚴格能力檢查（preflight 失敗時 raise ConfigurationError）"""
        self._strict_capability_check = enabled
        return self

    # ─────────────── Build ───────────────

    def build(self) -> "SystemControllerConfig":
        """建構 SystemControllerConfig"""
        return SystemControllerConfig(
            context_mappings=self._context_mappings,
            command_mappings=self._command_mappings,
            capability_context_mappings=self._cap_context_mappings,
            capability_command_mappings=self._cap_command_mappings,
            system_base=self._system_base,
            data_feed_mapping=self._data_feed_mapping,
            pv_max_history=self._pv_max_history,
            protection_rules=self._protection_rules,
            auto_stop_on_alarm=self._auto_stop_on_alarm,
            system_alarm_key=self._system_alarm_key,
            capacity_kva=self._capacity_kva,
            alarm_mode=self._alarm_mode,
            on_device_alarm=self._on_device_alarm,
            on_device_alarm_clear=self._on_device_alarm_clear,
            heartbeat_mappings=self._heartbeat_mappings,
            heartbeat_interval=self._heartbeat_interval,
            use_heartbeat_capability=self._use_heartbeat_capability,
            heartbeat_capability_mode=self._heartbeat_capability_mode,
            heartbeat_capability_constant_value=self._heartbeat_capability_constant_value,
            heartbeat_capability_increment_max=self._heartbeat_capability_increment_max,
            power_distributor=self._power_distributor,
            post_protection_processors=self._processors,
            runtime_params=self._runtime_params,
            capability_requirements=self._capability_requirements,
            strict_capability_check=self._strict_capability_check,
        )


class _OverrideState:
    """EventDrivenOverride 的內部狀態追蹤"""

    __slots__ = ("active", "deactivate_at")

    def __init__(self) -> None:
        self.active: bool = False
        self.deactivate_at: float | None = None


class SystemController(AsyncLifecycleMixin):
    """
    頂層系統控制器

    結合 ModeManager（模式管理）、ProtectionGuard（保護鏈）與既有整合元件，
    在不修改現有 GridControlLoop 的前提下，提供更高階的控制能力。

    內部流程：
    1. _build_context(): ContextBuilder.build() + 注入 system_alarm 旗標
    2. _on_command(command): ProtectionGuard.apply() → CommandRouter.route()
    3. _handle_auto_stop(): 告警啟用 → push stop override；告警解除 → pop

    使用範例::

        controller = SystemController(registry, config)
        controller.register_mode("pq", pq_strategy, ModePriority.SCHEDULE)
        await controller.set_base_mode("pq")
        async with controller:
            await asyncio.Event().wait()
    """

    def __init__(self, registry: DeviceRegistry, config: SystemControllerConfig) -> None:
        self._registry = registry
        self._config = config

        # 模式管理
        self._mode_manager = ModeManager(on_strategy_change=self._on_strategy_change)

        # 保護鏈
        self._protection_guard = ProtectionGuard(config.protection_rules)

        # PV 資料服務（可選）
        self._pv_service: PVDataService | None = None
        self._data_feed: DeviceDataFeed | None = None
        if config.data_feed_mapping is not None:
            self._pv_service = PVDataService(max_history=config.pv_max_history)
            self._data_feed = DeviceDataFeed(registry, config.data_feed_mapping, self._pv_service)

        # Context 建構器
        self._context_builder = ContextBuilder(
            registry,
            config.context_mappings,
            system_base=config.system_base,
            capability_mappings=config.capability_context_mappings or None,
            runtime_params=config.runtime_params,
        )

        # Command 路由器
        self._command_router = CommandRouter(
            registry,
            config.command_mappings,
            capability_mappings=config.capability_command_mappings or None,
        )

        # 系統指令編排器
        self._orchestrator = SystemCommandOrchestrator(registry)

        # 心跳服務（可選）
        self._heartbeat: HeartbeatService | None = None
        if config.heartbeat_mappings or config.use_heartbeat_capability:
            self._heartbeat = HeartbeatService(
                registry,
                mappings=config.heartbeat_mappings or None,
                interval=config.heartbeat_interval,
                use_capability=config.use_heartbeat_capability,
                mode=config.heartbeat_capability_mode,
                constant_value=config.heartbeat_capability_constant_value,
                increment_max=config.heartbeat_capability_increment_max,
            )

        # 策略執行器：context_provider 與 on_command 由本控制器管理
        self._executor = StrategyExecutor(
            context_provider=self._build_context,
            on_command=self._on_command,
        )

        # 自動停機狀態（向後相容）
        self._auto_stop_active = False
        self._cached_context: StrategyContext | None = None

        # 事件驅動 Override
        self._event_overrides: list[EventDrivenOverride] = []
        self._event_override_states: dict[str, _OverrideState] = {}

        # 設備級告警追蹤（per_device 模式）
        self._alarmed_devices: set[str] = set()

        # 背景任務
        self._run_task: asyncio.Task[None] | None = None

        # 註冊自動停機模式（通過 EventDrivenOverride 機制）
        if config.auto_stop_on_alarm:
            self._mode_manager.register(
                _AUTO_STOP_MODE,
                StopStrategy(),
                ModePriority.PROTECTION + 1,
                "Auto stop on system alarm",
            )
            self.register_event_override(AlarmStopOverride(name=_AUTO_STOP_MODE, alarm_key=config.system_alarm_key))

    # ---- 模式管理（委派 ModeManager）----

    def register_mode(self, name: str, strategy: Strategy, priority: int, description: str = "") -> None:
        """註冊模式，並驗證策略所需的 capabilities"""
        required = strategy.required_capabilities
        if required:
            for cap in required:
                devices = self._registry.get_devices_with_capability(cap)
                if not devices:
                    logger.warning(
                        f"Strategy '{strategy}' requires capability '{cap.name}' but no registered device has it."
                    )
        self._mode_manager.register(name, strategy, priority, description)

    async def set_base_mode(self, name: str | None) -> None:
        """設定基礎模式"""
        await self._mode_manager.set_base_mode(name)

    async def add_base_mode(self, name: str) -> None:
        """新增基礎模式（多 base mode 共存）"""
        await self._mode_manager.add_base_mode(name)

    async def remove_base_mode(self, name: str) -> None:
        """移除基礎模式"""
        await self._mode_manager.remove_base_mode(name)

    async def push_override(self, name: str) -> None:
        """推入 override 模式"""
        await self._mode_manager.push_override(name)

    async def pop_override(self, name: str) -> None:
        """移除 override 模式"""
        await self._mode_manager.pop_override(name)

    def register_event_override(self, override: EventDrivenOverride) -> None:
        """
        註冊事件驅動的 override

        Args:
            override: 實作 EventDrivenOverride 的實例，name 必須對應已註冊的模式
        """
        self._event_overrides.append(override)
        self._event_override_states[override.name] = _OverrideState()
        logger.debug(f"Event override registered: {override.name}")

    # ---- Preflight Check ----

    def preflight_check(self) -> list[str]:
        """驗證能力需求是否滿足

        Returns:
            不滿足的需求描述列表

        Raises:
            ConfigurationError: 當 strict_capability_check=True 且有不滿足的需求
        """
        if not self._config.capability_requirements:
            return []

        failures = self._registry.validate_capabilities(self._config.capability_requirements)
        if failures:
            for failure in failures:
                logger.warning(f"Preflight check: {failure}")
            if self._config.strict_capability_check:
                raise ConfigurationError(f"Preflight capability check failed: {'; '.join(failures)}")
        return failures

    # ---- 排程模式控制（ScheduleModeController 實作）----

    async def activate_schedule_mode(self, strategy: Strategy, *, description: str = "") -> None:
        """
        啟用排程模式（ScheduleModeController 實作）

        首次呼叫時註冊 ``__schedule__`` 模式並設為 base mode；
        後續呼叫更新策略，觸發 on_strategy_change。
        """
        mm = self._mode_manager
        if _SCHEDULE_MODE not in mm.registered_modes:
            mm.register(_SCHEDULE_MODE, strategy, ModePriority.SCHEDULE, description)
            await mm.add_base_mode(_SCHEDULE_MODE, source=SwitchSource.SCHEDULE)
        else:
            await mm.update_mode_strategy(
                _SCHEDULE_MODE,
                strategy,
                source=SwitchSource.SCHEDULE,
                description=description,
            )
            if _SCHEDULE_MODE not in mm.base_mode_names:
                await mm.add_base_mode(_SCHEDULE_MODE, source=SwitchSource.SCHEDULE)

    async def deactivate_schedule_mode(self) -> None:
        """
        停用排程模式（ScheduleModeController 實作）

        從 base mode 移除 ``__schedule__``。
        """
        mm = self._mode_manager
        if _SCHEDULE_MODE in mm.base_mode_names:
            await mm.remove_base_mode(_SCHEDULE_MODE)

    def trigger(self) -> None:
        """手動觸發策略執行"""
        self._executor.trigger()

    # ---- 生命週期 ----

    async def _on_start(self) -> None:
        """啟動系統控制器"""
        # Preflight: 驗證能力需求
        self.preflight_check()

        # 初始化 post-protection processors（async_init for MongoDB etc.）
        for proc in self._config.post_protection_processors:
            if hasattr(proc, "async_init"):
                await proc.async_init()
        if self._data_feed is not None:
            self._data_feed.attach()
        if self._heartbeat is not None:
            self._validate_heartbeat_points()
            await self._heartbeat.start()
        self._run_task = asyncio.create_task(self._executor.run())
        logger.info("SystemController started.")

    async def _on_stop(self) -> None:
        """停止系統控制器"""
        try:
            self._executor.stop()
            if self._run_task is not None:
                await self._run_task
                self._run_task = None
        finally:
            try:
                if self._heartbeat is not None:
                    await self._heartbeat.stop()
            finally:
                if self._data_feed is not None:
                    self._data_feed.detach()
        logger.info("SystemController stopped.")

    def _validate_heartbeat_points(self) -> None:
        """驗證心跳映射的 point_name 是否存在於目標設備（僅 warning，不中斷啟動）"""
        for mapping in self._config.heartbeat_mappings:
            if mapping.device_id is not None:
                device = self._registry.get_device(mapping.device_id)
                if device is not None and mapping.point_name not in device.all_point_names:
                    logger.warning(
                        "Heartbeat point '{}' not found on device '{}'.",
                        mapping.point_name,
                        mapping.device_id,
                    )
            elif mapping.trait is not None:
                devices = self._registry.get_devices_by_trait(mapping.trait)
                for device in devices:
                    if mapping.point_name not in device.all_point_names:
                        logger.warning(
                            "Heartbeat point '{}' not found on device '{}' (trait='{}').",
                            mapping.point_name,
                            device.device_id,
                            mapping.trait,
                        )

    # ---- 內部流程 ----

    def _build_context(self) -> StrategyContext:
        """建構策略上下文，注入 system_alarm 旗標"""
        context = self._context_builder.build()

        # 檢查所有設備的告警狀態
        has_alarm = any(dev.is_protected for dev in self._registry.all_devices)

        # per_device 模式不觸發系統級告警（由 _handle_device_alarms 逐設備處理）
        if self._config.alarm_mode == "per_device":
            context.extra[self._config.system_alarm_key] = False
        else:
            context.extra[self._config.system_alarm_key] = has_alarm

        self._cached_context = context
        return context

    async def _on_command(self, command: Command) -> None:
        """命令回呼：套用保護鏈 → post-protection processors → 路由到設備"""
        # 取得最近的 context（由 executor 在 _execute_strategy 中呼叫 _build_context 後產生）
        context = self._cached_context
        if context is None:
            context = self._build_context()

        # 套用保護鏈
        result = self._protection_guard.apply(command, context)
        protected_command = result.protected_command

        # 套用 post-protection processors（補償、日誌、審計等）
        final_command = protected_command
        for processor in self._config.post_protection_processors:
            try:
                final_command = await processor.process(final_command, context)
            except Exception:
                logger.opt(exception=True).warning(f"CommandProcessor {type(processor).__name__} failed, skipping")

        # 評估事件驅動 overrides（包含自動停機）
        if self._event_overrides:
            await self._evaluate_event_overrides(context)

        # 路由到設備（is_protected 設備由 CommandRouter 防禦性跳過）
        if self._config.power_distributor is not None and self._config.capability_command_mappings:
            snapshots = self._build_device_snapshots()
            per_device = self._config.power_distributor.distribute(final_command, snapshots)
            await self._command_router.route_per_device(final_command, per_device)
        else:
            await self._command_router.route(final_command)

        # 處理設備級告警（per_device 模式）
        if self._config.alarm_mode == "per_device":
            await self._handle_device_alarms()

    async def _evaluate_event_overrides(self, context: StrategyContext) -> None:
        """統一評估所有 EventDrivenOverride"""
        now = time.monotonic()

        for override in self._event_overrides:
            state = self._event_override_states.get(override.name)
            if state is None:
                continue

            should = override.should_activate(context)

            if should and not state.active:
                state.active = True
                state.deactivate_at = None
                try:
                    await self._mode_manager.push_override(override.name, source=SwitchSource.EVENT)
                    logger.warning(f"Event override activated: {override.name}")
                except (KeyError, ValueError):
                    pass

                # 向後相容：更新 _auto_stop_active
                if override.name == _AUTO_STOP_MODE:
                    self._auto_stop_active = True

            elif not should and state.active:
                if state.deactivate_at is None:
                    state.deactivate_at = now + override.cooldown_seconds

                if now >= state.deactivate_at:
                    state.active = False
                    state.deactivate_at = None
                    try:
                        await self._mode_manager.pop_override(override.name, source=SwitchSource.EVENT)
                        logger.info(f"Event override deactivated: {override.name}")
                    except KeyError:
                        pass

                    # 向後相容：更新 _auto_stop_active
                    if override.name == _AUTO_STOP_MODE:
                        self._auto_stop_active = False

    async def _handle_auto_stop(self, context: StrategyContext) -> None:
        """處理自動停機 override（已棄用，由 _evaluate_event_overrides 取代）"""
        await self._evaluate_event_overrides(context)

    async def _handle_device_alarms(self) -> None:
        """處理設備級告警：逐設備檢查，送關機指令或呼叫回呼"""
        for device in self._registry.all_devices:
            device_id = device.device_id
            if device.is_protected and device_id not in self._alarmed_devices:
                # 新增告警設備
                self._alarmed_devices.add(device_id)
                if self._config.on_device_alarm is not None:
                    await self._config.on_device_alarm(device)
                elif "stop" in getattr(device, "ACTIONS", {}):
                    await device.execute_action("stop")
                logger.warning(f"Device alarm activated: {device_id}")
            elif not device.is_protected and device_id in self._alarmed_devices:
                # 告警解除
                self._alarmed_devices.discard(device_id)
                if self._config.on_device_alarm_clear is not None:
                    await self._config.on_device_alarm_clear(device)
                logger.info(f"Device alarm cleared: {device_id}")

    def _build_device_snapshots(self) -> list[DeviceSnapshot]:
        """建構所有可用設備的狀態快照（供 PowerDistributor 使用）"""
        snapshots: list[DeviceSnapshot] = []
        for device in self._registry.all_devices:
            if not device.is_responsive or device.is_protected:
                continue
            metadata = self._registry.get_metadata(device.device_id)
            values = device.latest_values

            # 建構 capability slot → value 映射
            cap_values: dict[str, dict[str, Any]] = {}
            bindings = getattr(device, "capabilities", {})
            for cap_name, binding in bindings.items():
                slot_values: dict[str, Any] = {}
                for slot in binding.capability.read_slots:
                    point = binding.resolve(slot)
                    slot_values[slot] = values.get(point)
                cap_values[cap_name] = slot_values

            snapshots.append(
                DeviceSnapshot(
                    device_id=device.device_id,
                    metadata=metadata,
                    latest_values=values,
                    capabilities=cap_values,
                )
            )
        return snapshots

    def _resolve_strategy(self) -> Strategy | None:
        """根據 ModeManager 狀態解析當前應使用的策略"""
        mm = self._mode_manager
        if mm.active_override_names:
            return mm.effective_strategy  # 互斥 override
        base = mm.base_strategies
        if not base:
            return None
        if len(base) == 1:
            return base[0]
        if self._config.capacity_kva is not None:
            return CascadingStrategy(  # type: ignore[return-value]  # CascadingStrategy is structurally compatible
                base,
                CapacityConfig(self._config.capacity_kva),
                ExecutionConfig(mode=ExecutionMode.PERIODIC, interval_seconds=1),
            )
        return base[0]  # 無容量設定時 fallback 到最高優先權

    async def _on_strategy_change(self, old: Strategy | None, new: Strategy | None) -> None:
        """ModeManager 通知策略變更，統一走 _resolve_strategy"""
        resolved = self._resolve_strategy()
        logger.info(f"Strategy change: {old} -> {resolved}")
        await self._executor.set_strategy(resolved)

        # 依據策略的 suppress_heartbeat 控制心跳服務
        if self._heartbeat is not None and resolved is not None:
            if resolved.suppress_heartbeat:
                self._heartbeat.pause()
            else:
                self._heartbeat.resume()

    # ---- 唯讀屬性 ----

    @property
    def registry(self) -> DeviceRegistry:
        """設備查詢索引"""
        return self._registry

    @property
    def executor(self) -> StrategyExecutor:
        """策略執行器"""
        return self._executor

    @property
    def mode_manager(self) -> ModeManager:
        """模式管理器"""
        return self._mode_manager

    @property
    def protection_guard(self) -> ProtectionGuard:
        """保護鏈"""
        return self._protection_guard

    @property
    def effective_mode_name(self) -> str | None:
        """當前生效的模式名稱"""
        mode = self._mode_manager.effective_mode
        return mode.name if mode is not None else None

    @property
    def protection_status(self) -> ProtectionResult | None:
        """最近一次保護結果"""
        return self._protection_guard.last_result

    @property
    def pv_service(self) -> PVDataService | None:
        """PV 資料服務"""
        return self._pv_service

    @property
    def alarmed_device_ids(self) -> set[str]:
        """當前處於告警狀態的設備 ID 集合"""
        return set(self._alarmed_devices)

    def health(self) -> HealthReport:
        """取得系統健康報告（聚合所有設備）"""
        children = [dev.health() for dev in self._registry.all_devices]
        if all(c.status == HealthStatus.HEALTHY for c in children):
            status = HealthStatus.HEALTHY
        elif any(c.status == HealthStatus.UNHEALTHY for c in children):
            status = HealthStatus.UNHEALTHY
        else:
            status = HealthStatus.DEGRADED
        return HealthReport(
            status=status,
            component="system_controller",
            details={"mode": self.effective_mode_name, "alarmed": list(self._alarmed_devices)},
            children=children,
        )

    @property
    def context_builder(self) -> ContextBuilder:
        """Context 建構器"""
        return self._context_builder

    @property
    def command_router(self) -> CommandRouter:
        """Command 路由器"""
        return self._command_router

    @property
    def orchestrator(self) -> SystemCommandOrchestrator:
        """系統指令編排器"""
        return self._orchestrator

    @property
    def heartbeat(self) -> HeartbeatService | None:
        """心跳服務"""
        return self._heartbeat

    @property
    def event_overrides(self) -> list[EventDrivenOverride]:
        """已註冊的事件驅動 override 列表"""
        return list(self._event_overrides)

    @property
    def auto_stop_active(self) -> bool:
        """自動停機是否啟動"""
        return self._auto_stop_active

    @property
    def config(self) -> SystemControllerConfig:
        """系統控制器配置"""
        return self._config

    @property
    def is_running(self) -> bool:
        """是否正在執行"""
        return self._run_task is not None and not self._run_task.done()



# =============== Equipment Alarm - State ===============
#
# 告警狀態管理
#
# 提供告警狀態持久化與遲滯處理

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum

from .definition import AlarmDefinition, AlarmLevel


class AlarmEventType(Enum):
    """告警事件類型"""

    TRIGGERED = "triggered"
    CLEARED = "cleared"


@dataclass
class AlarmEvent:
    """告警事件"""

    event_type: AlarmEventType
    alarm: AlarmDefinition
    timestamp: datetime = field(default_factory=lambda: datetime.now(timezone.utc))


@dataclass
class AlarmState:
    """
    告警狀態 - 可變

    Attributes:
        definition: 告警定義
        is_active: 是否啟用中
        activate_count: 連續觸發次數
        clear_count: 連續解除次數
        activated_at: 本次告警啟用時間
        cleared_at: 本次告警清除時間
        last_triggered_at: 最後條件成立時間（用於追蹤）
    """

    definition: AlarmDefinition
    is_active: bool = False
    activate_count: int = 0
    clear_count: int = 0
    activated_at: datetime | None = None
    cleared_at: datetime | None = None
    last_triggered_at: datetime | None = None

    @property
    def duration(self) -> float | None:
        """
        告警持續時間（秒）

        Returns:
            - 告警已清除：返回 activated_at 到 cleared_at 的秒數
            - 告警進行中：返回 activated_at 到現在的秒數
            - 尚未啟用過：返回 None
        """
        if self.activated_at is None:
            return None
        end_time = self.cleared_at if self.cleared_at else datetime.now(timezone.utc)
        return (end_time - self.activated_at).total_seconds()

    def update(self, is_triggered: bool) -> AlarmEvent | None:
        """
        更新告警狀態

        根據遲滯設定決定是否真正觸發/解除告警。

        Args:
            is_triggered: 本次評估結果

        Returns:
            告警事件（如果狀態變化），否則 None
        """
        hysteresis = self.definition.hysteresis
        now = datetime.now(timezone.utc)

        if is_triggered:
            # 觸發中
            self.activate_count += 1
            self.clear_count = 0
            self.last_triggered_at = now

            if not self.is_active and self.activate_count >= hysteresis.activate_threshold:
                # 達到觸發閾值，啟用告警
                self.is_active = True
                self.activated_at = now
                self.cleared_at = None  # 新週期開始，清除上次的結束時間
                return AlarmEvent(AlarmEventType.TRIGGERED, self.definition, now)
        else:
            # 未觸發
            self.clear_count += 1
            self.activate_count = 0

            if self.is_active and self.clear_count >= hysteresis.clear_threshold:
                # 達到解除閾值，清除告警
                self.is_active = False
                self.cleared_at = now
                return AlarmEvent(AlarmEventType.CLEARED, self.definition, now)

        return None

    def force_clear(self) -> AlarmEvent | None:
        """強制清除告警"""
        if self.is_active:
            now = datetime.now(timezone.utc)
            self.is_active = False
            self.activate_count = 0
            self.clear_count = 0
            self.cleared_at = now
            return AlarmEvent(AlarmEventType.CLEARED, self.definition, now)
        return None


class AlarmStateManager:
    """
    告警狀態管理器

    管理所有告警的狀態，提供持久化與遲滯處理。
    重要：讀取失敗時保持現有狀態，不會清除告警。
    """

    def __init__(self) -> None:
        self._states: dict[str, AlarmState] = {}

    def register_alarm(self, alarm: AlarmDefinition) -> AlarmState:
        """
        註冊告警定義

        Args:
            alarm: 告警定義

        Returns:
            AlarmState: 新建或已存在的告警狀態

        Raises:
            KeyError: 告警代碼已存在
        """
        if alarm.code in self._states:
            raise KeyError(f"告警 '{alarm.code}' 已註冊")
        else:
            self._states[alarm.code] = AlarmState(definition=alarm)
        return self._states[alarm.code]

    def register_alarms(self, alarms: list[AlarmDefinition]) -> list[AlarmState]:
        """
        批量註冊告警定義

        Args:
            alarms: 告警定義列表

        Returns:
            list[AlarmState]: 成功註冊的告警狀態列表

        Raises:
            KeyError: 存在重複的告警代碼
        """
        duplicates = [a.code for a in alarms if a.code in self._states]
        if duplicates:
            raise KeyError(f"以下告警代碼已存在: {duplicates}")

        return [self.register_alarm(alarm) for alarm in alarms]

    def update(self, evaluations: dict[str, bool]) -> list[AlarmEvent]:
        """
        更新告警狀態

        Args:
            evaluations: {告警代碼: 是否觸發} 字典

        Returns:
            告警事件列表（狀態變化的告警）
        """
        events: list[AlarmEvent] = []

        for code, is_triggered in evaluations.items():
            if code not in self._states:
                continue

            state = self._states[code]
            event = state.update(is_triggered=is_triggered)
            if event:
                events.append(event)

        return events

    def clear_alarm(self, code: str) -> AlarmEvent | None:
        """
        強制清除告警

        Args:
            code: 告警代碼

        Returns:
            AlarmEvent: 清除事件，如果告警不存在則返回 None
        """
        if code not in self._states:
            return None

        state = self._states[code]
        return state.force_clear()

    def get_active_alarms(self) -> list[AlarmState]:
        """
        取得所有啟用中的告警

        Returns:
            list[AlarmState]: 所有啟用中的告警狀態
        """
        return [state for state in self._states.values() if state.is_active]

    def get_state(self, code: str) -> AlarmState | None:
        """
        取得特定告警的狀態

        Args:
            code: 告警代碼

        Returns:
            AlarmState: 告警狀態，如果告警不存在則返回 None
        """
        return self._states.get(code)

    def has_protection_alarm(self) -> bool:
        """
        檢查是否存在保護性告警

        Returns:
            bool: 如果存在保護性告警則返回 True，否則返回 False
        """
        return any(state.is_active and state.definition.level == AlarmLevel.ALARM for state in self._states.values())

    def export_states(self) -> dict[str, AlarmState]:
        """
        匯出所有告警狀態的 shallow copy

        Returns:
            {告警代碼: AlarmState} 字典
        """
        return dict(self._states)

    def import_states(self, states: dict[str, AlarmState]) -> None:
        """
        匯入告警狀態

        對已註冊的 code，用舊 state 覆蓋新 state 的計數和時間欄位。

        Args:
            states: 要匯入的告警狀態
        """
        for code, old_state in states.items():
            if code in self._states:
                new_state = self._states[code]
                new_state.is_active = old_state.is_active
                new_state.activate_count = old_state.activate_count
                new_state.clear_count = old_state.clear_count
                new_state.activated_at = old_state.activated_at
                new_state.cleared_at = old_state.cleared_at
                new_state.last_triggered_at = old_state.last_triggered_at

    def reset(self) -> None:
        """
        重置所有告警狀態
        """
        for state in self._states.values():
            state.force_clear()


__all__ = [
    "AlarmEventType",
    "AlarmEvent",
    "AlarmState",
    "AlarmStateManager",
]



"""
CSP GUI - Web-based Runtime Control Panel

Browser-based monitoring and control panel for SystemController.

Optional dependency: pip install csp0924_lib[gui]

Usage::

    from csp_lib.integration import SystemController
    from csp_lib.gui import create_app, GUIConfig

    app = create_app(controller)

    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8080)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .app import create_app
from .config import GUIConfig

if TYPE_CHECKING:
    pass

__all__ = [
    "create_app",
    "GUIConfig",
]



# =============== Manager - State ===============
#
# 狀態同步管理模組
#
# 提供設備狀態同步至 Redis 功能：
#   - StateSyncManager: 訂閱設備事件並同步至 Redis + Pub/Sub

from .config import StateSyncConfig
from .sync import StateSyncManager

__all__ = [
    "StateSyncConfig",
    "StateSyncManager",
]



# =============== Manager Alarm - Repository ===============
#
# 告警資料存取層
#
# 提供告警資料的 CRUD 操作：
#   - AlarmRepository: 資料存取介面 (Protocol)
#   - MongoAlarmRepository: MongoDB 實作
#
# 設計原則：
#   - 依賴倒置：業務層依賴 Protocol，不依賴具體實作
#   - 單一職責：僅負責資料存取，不包含業務邏輯

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Protocol, runtime_checkable

from .schema import AlarmRecord, AlarmStatus

if TYPE_CHECKING:
    from motor.motor_asyncio import AsyncIOMotorDatabase


@runtime_checkable
class AlarmRepository(Protocol):
    """
    告警 Repository 介面

    定義告警資料存取的標準介面，遵循依賴倒置原則。
    業務層應依賴此 Protocol，而非具體實作。
    """

    async def health_check(self) -> bool:
        """檢查 Repository 連線是否正常"""
        ...

    async def upsert(self, record: AlarmRecord) -> tuple[str, bool]:
        """新增或更新告警記錄"""
        ...

    async def resolve(self, alarm_key: str, resolved_at: datetime) -> bool:
        """解除告警（更新 resolved_timestamp 與 status）"""
        ...

    async def get_active_alarms(self) -> list[AlarmRecord]:
        """取得所有進行中的告警"""
        ...

    async def get_active_by_device(self, device_id: str) -> list[AlarmRecord]:
        """取得指定設備的進行中告警"""
        ...


class MongoAlarmRepository:
    """
    MongoDB 告警 Repository 實作

    使用 Motor 非同步驅動實作 AlarmRepository 介面。
    支援告警的 CRUD 操作與索引管理。

    Attributes:
        COLLECTION_NAME: 預設 collection 名稱
    """

    COLLECTION_NAME = "alarms"

    def __init__(self, db: AsyncIOMotorDatabase, collection_name: str = COLLECTION_NAME) -> None:
        """
        初始化 MongoDB Repository

        Args:
            db: Motor 非同步資料庫連線
            collection_name: Collection 名稱（預設 "alarms"）
        """
        self._db = db
        self._collection = db[collection_name]

    async def health_check(self) -> bool:
        """
        檢查 MongoDB 連線是否正常

        Returns:
            bool: True 表示連線正常
        """
        try:
            await self._db.command("ping")
            return True
        except Exception:
            return False

    async def ensure_indexes(self) -> None:
        """
        建立資料庫索引

        應於應用程式啟動時呼叫一次。建立的索引包括：
        - alarm_key: 唯一鍵查詢
        - device_id: 設備篩選
        - status: 狀態篩選
        - (device_id, status): 複合查詢
        - timestamp: 時間排序
        """
        from pymongo import IndexModel

        await self._collection.create_indexes(
            [
                IndexModel([("alarm_key", 1)]),
                IndexModel([("device_id", 1)]),
                IndexModel([("status", 1)]),
                IndexModel([("device_id", 1), ("status", 1)]),
                IndexModel([("timestamp", -1)]),
            ]
        )

    async def upsert(self, record: AlarmRecord) -> tuple[str, bool]:
        """
        新增告警記錄

        若已存在相同 alarm_key 且狀態為 ACTIVE 的告警，則跳過新增。
        此設計避免重複寫入同一告警事件。

        Args:
            record: 告警記錄

        Returns:
            tuple[str, bool]: (告警 ID, 是否為新增)
                - 若為既有告警：返回既有 _id，is_new=False
                - 若為新告警：返回新 _id，is_new=True
        """
        existing = await self._collection.find_one(
            {
                "alarm_key": record.alarm_key,
                "status": AlarmStatus.ACTIVE.value,
            }
        )

        if existing:
            return existing["_id"], False
        result = await self._collection.insert_one(record.to_document())
        return result.inserted_id, True

    async def resolve(self, alarm_key: str, resolved_at: datetime) -> bool:
        """
        解除告警

        將指定告警的狀態從 ACTIVE 更新為 RESOLVED，
        並記錄解除時間。僅更新狀態為 ACTIVE 的告警。

        Args:
            alarm_key: 告警唯一鍵
            resolved_at: 解除時間

        Returns:
            bool: 是否成功解除（True 表示有更新，False 表示無對應告警）
        """
        result = await self._collection.update_one(
            {
                "alarm_key": alarm_key,
                "status": AlarmStatus.ACTIVE.value,
            },
            {
                "$set": {
                    "status": AlarmStatus.RESOLVED.value,
                    "resolved_timestamp": resolved_at,
                }
            },
        )
        return result.modified_count == 1

    async def get_active_alarms(self) -> list[AlarmRecord]:
        """
        取得所有進行中的告警

        Returns:
            list[AlarmRecord]: 所有狀態為 ACTIVE 的告警清單
        """
        cursor = self._collection.find(
            {
                "status": AlarmStatus.ACTIVE.value,
            }
        )
        return [AlarmRecord.from_document(doc) async for doc in cursor]

    async def get_active_by_device(self, device_id: str) -> list[AlarmRecord]:
        """
        取得指定設備的進行中告警

        Args:
            device_id: 設備識別碼

        Returns:
            list[AlarmRecord]: 該設備所有狀態為 ACTIVE 的告警清單
        """
        cursor = self._collection.find(
            {
                "device_id": device_id,
                "status": AlarmStatus.ACTIVE.value,
            }
        )
        return [AlarmRecord.from_document(doc) async for doc in cursor]



# =============== Integration Distributed - Subscriber ===============
#
# 設備狀態訂閱器
#
# 從 Redis 輪詢遠端設備的狀態、連線狀態與告警：
#   - DeviceStateSubscriber: 分散式設備狀態輪詢器

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

from csp_lib.core import AsyncLifecycleMixin, get_logger

if TYPE_CHECKING:
    from csp_lib.redis import RedisClient

    from .config import DistributedConfig

logger = get_logger(__name__)


class DeviceStateSubscriber(AsyncLifecycleMixin):
    """
    設備狀態訂閱器

    從 Redis 輪詢遠端站台設備的狀態，提供與 ClusterStateSubscriber
    相同的 device_states 介面，供 VirtualContextBuilder 使用。

    Redis Key 結構（由遠端 StateSyncManager 發佈）：
        - device:{device_id}:state - Hash，所有點位最新值
        - device:{device_id}:online - String，"1" 或 "0"
        - device:{device_id}:alarms - Set，活躍告警 codes

    Usage::

        subscriber = DeviceStateSubscriber(config, redis_client)
        async with subscriber:
            states = subscriber.device_states
            online = subscriber.device_online
    """

    def __init__(
        self,
        config: DistributedConfig,
        redis_client: RedisClient,
    ) -> None:
        self._config = config
        self._redis = redis_client
        self._device_states: dict[str, dict[str, Any]] = {}
        self._device_online: dict[str, bool] = {}
        self._device_alarms: dict[str, set[str]] = {}
        self._task: asyncio.Task | None = None
        self._stop_event = asyncio.Event()

    @property
    def device_states(self) -> dict[str, dict[str, Any]]:
        """設備狀態快取（device_id -> latest_values dict）"""
        return self._device_states

    @property
    def device_online(self) -> dict[str, bool]:
        """設備連線狀態（device_id -> is_online）"""
        return self._device_online

    @property
    def device_alarms(self) -> dict[str, set[str]]:
        """設備活躍告警（device_id -> alarm codes）"""
        return self._device_alarms

    async def _on_start(self) -> None:
        """啟動輪詢迴圈"""
        self._stop_event.clear()
        self._task = asyncio.create_task(self._poll_loop())
        logger.info("DeviceStateSubscriber started.")

    async def _on_stop(self) -> None:
        """停止輪詢"""
        self._stop_event.set()
        if self._task is not None and not self._task.done():
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
            self._task = None
        logger.info("DeviceStateSubscriber stopped.")

    async def _poll_loop(self) -> None:
        """輪詢主迴圈"""
        interval = self._config.poll_interval
        while not self._stop_event.is_set():
            try:
                await self._poll_all()
            except asyncio.CancelledError:
                return
            except Exception:
                logger.opt(exception=True).warning("Failed to poll device states")

            try:
                await asyncio.wait_for(self._stop_event.wait(), timeout=interval)
                return
            except asyncio.TimeoutError:
                pass

    async def _poll_all(self) -> None:
        """讀取所有設備狀態"""
        for device_id in self._config.all_device_ids:
            # State (HGETALL)
            try:
                state = await self._redis.hgetall(f"device:{device_id}:state")
                if state:
                    self._device_states[device_id] = state
                else:
                    self._device_states.pop(device_id, None)
            except Exception:
                logger.warning(f"Failed to read state for device {device_id}")

            # Online (GET)
            try:
                online_raw = await self._redis.get(f"device:{device_id}:online")
                self._device_online[device_id] = online_raw == "1"
            except Exception:
                logger.warning(f"Failed to read online status for device {device_id}")
                self._device_online[device_id] = False

            # Alarms (SMEMBERS)
            try:
                alarms = await self._redis.smembers(f"device:{device_id}:alarms")
                self._device_alarms[device_id] = alarms
            except Exception:
                logger.warning(f"Failed to read alarms for device {device_id}")
                self._device_alarms[device_id] = set()


__all__ = [
    "DeviceStateSubscriber",
]



# =============== Integration Distributed Module ===============
#
# 分散式控制模組
#
# 支援多機分散式設備控制架構：
#   - DistributedConfig / RemoteSiteConfig: 分散式配置
#   - DeviceStateSubscriber: 從 Redis 輪詢遠端設備狀態
#   - RemoteCommandRouter: 透過 Redis 發送指令到遠端站台
#   - DistributedController: Controller 端主編排器
#   - RemoteSiteRunner: 遠端站台設備端執行器

from .command_router import RemoteCommandRouter
from .config import DistributedConfig, RemoteSiteConfig
from .controller import DistributedController
from .site_runner import RemoteSiteRunner
from .subscriber import DeviceStateSubscriber

__all__ = [
    # Config
    "DistributedConfig",
    "RemoteSiteConfig",
    # Subscriber
    "DeviceStateSubscriber",
    # Command Router
    "RemoteCommandRouter",
    # Controller
    "DistributedController",
    # Site Runner
    "RemoteSiteRunner",
]



# =============== Modbus Gateway Module ===============
#
# Modbus Gateway 模組
#
# 提供 Modbus TCP 閘道伺服器，讓 PCS/BMS 等設備以 Modbus slave
# 身份對外暴露暫存器，供 EMS 或 SCADA 讀寫。
#
# Usage:
#     from csp_lib.modbus_gateway import (
#         # Errors
#         GatewayError,
#         RegisterConflictError,
#         WriteRejectedError,
#         # Config
#         RegisterType,
#         GatewayRegisterDef,
#         WatchdogConfig,
#         GatewayServerConfig,
#         # Protocol
#         WriteValidator,
#         WriteHook,
#         WriteRule,
#         DataSyncSource,
#         UpdateRegisterCallback,
#         # Rules
#         RangeRule,
#         AllowedValuesRule,
#         StepRule,
#         CompositeRule,
#         # Watchdog
#         CommunicationWatchdog,
#         # Validators
#         AddressWhitelistValidator,
#     )

from .config import (
    GatewayRegisterDef,
    GatewayServerConfig,
    RegisterType,
    WatchdogConfig,
    WriteRule,
)
from .errors import (
    GatewayError,
    RegisterConflictError,
    WriteRejectedError,
)
from .hooks import CallbackHook, RedisPublishHook, StatePersistHook
from .protocol import (
    DataSyncSource,
    UpdateRegisterCallback,
    WriteHook,
    WriteValidator,
)
from .register_map import GatewayRegisterMap
from .server import ModbusGatewayServer
from .sync_sources import PollingCallbackSource, RedisSubscriptionSource
from .validators import AddressWhitelistValidator
from .watchdog import CommunicationWatchdog

__all__ = [
    # Errors
    "GatewayError",
    "RegisterConflictError",
    "WriteRejectedError",
    # Config
    "RegisterType",
    "GatewayRegisterDef",
    "WatchdogConfig",
    "GatewayServerConfig",
    # Protocol
    "WriteValidator",
    "WriteHook",
    "WriteRule",
    "DataSyncSource",
    "UpdateRegisterCallback",
    # Core
    "GatewayRegisterMap",
    "ModbusGatewayServer",
    # Watchdog
    "CommunicationWatchdog",
    # Validators
    "AddressWhitelistValidator",
    # Hooks
    "RedisPublishHook",
    "CallbackHook",
    "StatePersistHook",
    # Sync Sources
    "RedisSubscriptionSource",
    "PollingCallbackSource",
]



# =============== Equipment Template Module ===============
#
# 設備範本模組匯出
#
# 提供可重用的設備模型範本與工廠：
#   - EquipmentTemplate: 設備模型定義
#   - PointOverride: 點位覆寫定義
#   - DeviceFactory: 設備工廠

from .definition import EquipmentTemplate, PointOverride
from .factory import DeviceFactory

__all__ = [
    "EquipmentTemplate",
    "PointOverride",
    "DeviceFactory",
]



# =============== Bypass Strategy ===============
#
# 旁路策略：什麼都不做，用於手動控制模式

from __future__ import annotations

from csp_lib.controller.core import Command, ExecutionConfig, ExecutionMode, Strategy, StrategyContext


class BypassStrategy(Strategy):
    """
    旁路策略 (Bypass Mode)

    完全不發送任何指令，用於手動控制模式。
    GridController 的 command loop 會跳過此策略，
    使用者可透過外部方式直接控制設備。

    使用情境：
        - 手動調試設備
        - 臨時接管控制權
        - 維護模式
    """

    @property
    def suppress_heartbeat(self) -> bool:
        """旁路模式暫停心跳，讓設備知道控制器已釋放控制權"""
        return True

    @property
    def execution_config(self) -> ExecutionConfig:
        # TRIGGERED 模式，不會主動執行
        return ExecutionConfig(mode=ExecutionMode.TRIGGERED, interval_seconds=1)

    def execute(self, context: StrategyContext) -> Command:
        # 維持上一次的命令狀態
        return context.last_command

    def __str__(self) -> str:
        return "BypassStrategy()"



# =============== Modbus Server - Curve Behavior ===============
#
# 曲線驅動行為：整合現有 CurveRegistry

from __future__ import annotations

import time
from typing import Iterator

from csp_lib.equipment.simulation.curve import CurvePoint, CurveProvider


class CurveBehavior:
    """
    曲線驅動行為

    使用 CurveProvider 按時間驅動數值變化。
    每個 CurvePoint 指定一個 value 和 duration，
    在 duration 內持續輸出該 value。
    """

    def __init__(self, curve_provider: CurveProvider, default_value: float = 0.0) -> None:
        self._curve_provider = curve_provider
        self._default_value = default_value
        self._current_value = default_value
        self._iterator: Iterator[CurvePoint] | None = None
        self._curve_name: str | None = None
        self._current_point: CurvePoint | None = None
        self._point_end_time: float | None = None

    @property
    def current_value(self) -> float:
        return self._current_value

    @property
    def is_running(self) -> bool:
        return self._iterator is not None

    def start_curve(self, name: str) -> bool:
        """啟動指定曲線"""
        curve = self._curve_provider.get_curve(name)
        if curve is None:
            return False
        self._iterator = curve
        self._curve_name = name
        return self._advance()

    def stop_curve(self) -> None:
        """停止曲線"""
        self._iterator = None
        self._curve_name = None
        self._current_point = None
        self._point_end_time = None
        self._current_value = self._default_value

    def update(self) -> float:
        """
        更新曲線狀態

        Returns:
            當前值（曲線值或 default_value）
        """
        if self._iterator is None:
            return self._current_value

        now = time.monotonic()
        if self._point_end_time is not None and now >= self._point_end_time:
            if not self._advance():
                return self._current_value

        if self._current_point is not None:
            self._current_value = self._current_point.value

        return self._current_value

    def _advance(self) -> bool:
        """前進到下一個曲線點"""
        if self._iterator is None:
            return False
        try:
            point = next(self._iterator)
            self._current_point = point
            self._point_end_time = time.monotonic() + point.duration
            self._current_value = point.value
            return True
        except StopIteration:
            self.stop_curve()
            return False

    def reset(self) -> None:
        """重置到初始狀態"""
        self.stop_curve()


__all__ = ["CurveBehavior"]
